#!/usr/bin/env python
# encoding: utf8
#
# Copyright (C) 2010-2019 Pytroll
#
# Authors:
#    Esben S. Nielsen
#    Thomas Lavergne
#    Adam Dybbroe <adam.dybbroe@smhi.se>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Utility functions for quick and easy display."""

from __future__ import absolute_import

import numpy as np

try:
    from pyresample.utils import cartopy  # noqa

    BASEMAP_NOT_CARTOPY = False
except ImportError:
    BASEMAP_NOT_CARTOPY = True


def ellps2axis(ellps_name):
    """Get semi-major and semi-minor axis from ellipsis definition.

    Parameters
    ---------
    ellps_name : str
        Standard name of ellipsis

    Returns
    -------
    (a, b) : semi-major and semi-minor axis
    """
    ellps = {'helmert': {'a': 6378200.0, 'b': 6356818.1696278909},
             'intl': {'a': 6378388.0, 'b': 6356911.9461279465},
             'merit': {'a': 6378137.0, 'b': 6356752.2982159676},
             'wgs72': {'a': 6378135.0, 'b': 6356750.5200160937},
             'sphere': {'a': 6370997.0, 'b': 6370997.0},
             'clrk66': {'a': 6378206.4000000004, 'b': 6356583.7999999998},
             'nwl9d': {'a': 6378145.0, 'b': 6356759.7694886839},
             'lerch': {'a': 6378139.0, 'b': 6356754.2915103417},
             'evrstss': {'a': 6377298.5559999999, 'b': 6356097.5503008962},
             'evrst30': {'a': 6377276.3449999997, 'b': 6356075.4131402401},
             'mprts': {'a': 6397300.0, 'b': 6363806.2827225132},
             'krass': {'a': 6378245.0, 'b': 6356863.0187730473},
             'walbeck': {'a': 6376896.0, 'b': 6355834.8466999996},
             'kaula': {'a': 6378163.0, 'b': 6356776.9920869097},
             'wgs66': {'a': 6378145.0, 'b': 6356759.7694886839},
             'evrst56': {'a': 6377301.2429999998, 'b': 6356100.2283681016},
             'new_intl': {'a': 6378157.5, 'b': 6356772.2000000002},
             'airy': {'a': 6377563.3959999997, 'b': 6356256.9100000001},
             'bessel': {'a': 6377397.1550000003, 'b': 6356078.9628181886},
             'seasia': {'a': 6378155.0, 'b': 6356773.3205000004},
             'aust_sa': {'a': 6378160.0, 'b': 6356774.7191953054},
             'wgs84': {'a': 6378137.0, 'b': 6356752.3142451793},
             'hough': {'a': 6378270.0, 'b': 6356794.3434343431},
             'wgs60': {'a': 6378165.0, 'b': 6356783.2869594367},
             'engelis': {'a': 6378136.0499999998, 'b': 6356751.3227215428},
             'apl4.9': {'a': 6378137.0, 'b': 6356751.796311819},
             'andrae': {'a': 6377104.4299999997, 'b': 6355847.4152333336},
             'sgs85': {'a': 6378136.0, 'b': 6356751.301568781},
             'delmbr': {'a': 6376428.0, 'b': 6355957.9261637237},
             'fschr60m': {'a': 6378155.0, 'b': 6356773.3204827355},
             'iau76': {'a': 6378140.0, 'b': 6356755.2881575283},
             'plessis': {'a': 6376523.0, 'b': 6355863.0},
             'cpm': {'a': 6375738.7000000002, 'b': 6356666.221912113},
             'fschr68': {'a': 6378150.0, 'b': 6356768.3372443849},
             'mod_airy': {'a': 6377340.1890000002, 'b': 6356034.4460000005},
             'grs80': {'a': 6378137.0, 'b': 6356752.3141403561},
             'bess_nam': {'a': 6377483.8650000002, 'b': 6356165.3829663256},
             'fschr60': {'a': 6378166.0, 'b': 6356784.2836071067},
             'clrk80': {'a': 6378249.1449999996, 'b': 6356514.9658284895},
             'evrst69': {'a': 6377295.6639999999, 'b': 6356094.6679152036},
             'grs67': {'a': 6378160.0, 'b': 6356774.5160907144},
             'evrst48': {'a': 6377304.0630000001, 'b': 6356103.0389931547}}
    try:
        ellps_axis = ellps[ellps_name.lower()]
        a = ellps_axis['a']
        b = ellps_axis['b']
    except KeyError:
        raise ValueError(('Could not determine semi-major and semi-minor axis '
                          'of specified ellipsis %s') % ellps_name)
    return a, b


def area_def2basemap(area_def, **kwargs):
    """Get Basemap object from an AreaDefinition object.

    Parameters
    ---------
    area_def : object
        geometry.AreaDefinition object
    **kwargs: Keyword arguments
        Additional initialization arguments for Basemap

    Returns
    -------
    bmap : Basemap object
    """
    import warnings
    warnings.warn("Basemap is no longer maintained. Please switch to cartopy "
                  "by using 'area_def.to_cartopy_crs()'. See the pyresample "
                  "documentation for more details.", DeprecationWarning)

    from mpl_toolkits.basemap import Basemap
    try:
        a, b = ellps2axis(area_def.proj_dict['ellps'])
        rsphere = (a, b)
    except KeyError:
        try:
            a = float(area_def.proj_dict['a'])
            try:
                b = float(area_def.proj_dict['b'])
                rsphere = (a, b)
            except KeyError:
                rsphere = a
        except KeyError:
            # Default to WGS84 ellipsoid
            a, b = ellps2axis('wgs84')
            rsphere = (a, b)

    # Add projection specific basemap args to args passed to function
    basemap_args = kwargs
    basemap_args['rsphere'] = rsphere

    if area_def.proj_dict['proj'] in ('ortho', 'geos', 'nsper'):
        llcrnrx, llcrnry, urcrnrx, urcrnry = area_def.area_extent
        basemap_args['llcrnrx'] = llcrnrx
        basemap_args['llcrnry'] = llcrnry
        basemap_args['urcrnrx'] = urcrnrx
        basemap_args['urcrnry'] = urcrnry
    else:
        llcrnrlon, llcrnrlat, urcrnrlon, urcrnrlat = area_def.area_extent_ll
        basemap_args['llcrnrlon'] = llcrnrlon
        basemap_args['llcrnrlat'] = llcrnrlat
        basemap_args['urcrnrlon'] = urcrnrlon
        basemap_args['urcrnrlat'] = urcrnrlat

    if area_def.proj_dict['proj'] == 'eqc':
        basemap_args['projection'] = 'cyl'
    else:
        basemap_args['projection'] = area_def.proj_dict['proj']

    # Try adding potentially remaining args
    for key in ('lon_0', 'lat_0', 'lon_1', 'lat_1', 'lon_2', 'lat_2',
                'lat_ts'):
        try:
            basemap_args[key] = float(area_def.proj_dict[key])
        except KeyError:
            pass

    return Basemap(**basemap_args)


def _basemap_get_quicklook(area_def, data, vmin=None, vmax=None,
                           label='Variable (units)', num_meridians=45,
                           num_parallels=10, coast_res='110m', cmap='RdBu_r'):
    """Doing quicklook image plots with Basemap."""
    if area_def.shape != data.shape:
        raise ValueError('area_def shape %s does not match data shape %s' %
                         (list(area_def.shape), list(data.shape)))
    import matplotlib.pyplot as plt
    bmap = area_def2basemap(area_def, resolution=coast_res)
    bmap.drawcoastlines()
    if num_meridians > 0:
        bmap.drawmeridians(np.arange(-180, 180, num_meridians))
    if num_parallels > 0:
        bmap.drawparallels(np.arange(-90, 90, num_parallels))
    if not (np.ma.isMaskedArray(data) and data.mask.all()):
        col = bmap.imshow(data, origin='upper', vmin=vmin, vmax=vmax, cmap=cmap)
        plt.colorbar(col, shrink=0.5, pad=0.05).set_label(label)
    return plt


def _translate_coast_resolution_to_cartopy(coast_res):
    """Translate the coast resolution argument between cartopy and basemap notation."""
    bmap_to_cartopy_res = {
        'c': '110m',
        'l': '110m',
        'i': '50m',
        'h': '10m',
        'f': '10m'
    }

    if BASEMAP_NOT_CARTOPY:
        if coast_res.endswith('m'):
            _rev_map = {v: k for k, v in bmap_to_cartopy_res.items()}
            coast_res = _rev_map[coast_res]
        return coast_res, False

    if coast_res and coast_res not in ['110m', '50m', '10m']:
        import warnings
        warnings.warn("'coast_res' should be either '110m', '50m', '10m'.")
        coast_res = {
            'c': '110m',
            'l': '110m',
            'i': '50m',
            'h': '10m',
            'f': '10m'
        }[coast_res]

    return coast_res, True


def _add_gridlines(axes, nmeridians, nparallels):
    """Add gridlines: meridians and parallels onto the plot."""
    from matplotlib import ticker as mticker

    gl = axes.gridlines()
    if nmeridians:
        gl.xlocator = mticker.FixedLocator(np.arange(-180, 180 + nmeridians, nmeridians))
    else:
        gl.xlines = False
    if nparallels:
        gl.ylocator = mticker.FixedLocator(np.arange(-90, 90 + nparallels, nparallels))
    else:
        gl.ylines = False

    return gl


def _get_quicklook(area_def, data, vmin=None, vmax=None,
                   label='Variable (units)', num_meridians=45,
                   num_parallels=10, coast_res='110m', cmap='RdBu_r'):
    """Get default cartopy matplotlib plot."""
    import matplotlib.pyplot as plt

    coast_res, is_cartopy = _translate_coast_resolution_to_cartopy(coast_res)
    if not is_cartopy:
        return _basemap_get_quicklook(
            area_def, data, vmin, vmax, label, num_meridians,
            num_parallels, coast_res=coast_res, cmap=cmap)

    if area_def.shape != data.shape:
        raise ValueError('area_def shape %s does not match data shape %s' %
                         (list(area_def.shape), list(data.shape)))

    crs = area_def.to_cartopy_crs()
    ax = plt.axes(projection=crs)
    ax.coastlines(resolution=coast_res)
    ax.set_global()

    if num_meridians or num_parallels:
        _ = _add_gridlines(ax, num_meridians, num_parallels)

    if not (np.ma.isMaskedArray(data) and data.mask.all()):
        col = ax.imshow(data, transform=crs, extent=crs.bounds,
                        origin='upper', vmin=vmin, vmax=vmax, cmap=cmap)
        plt.colorbar(col, shrink=0.5, pad=0.05).set_label(label)
    return plt


def show_quicklook(area_def, data, vmin=None, vmax=None,
                   label='Variable (units)', num_meridians=45,
                   num_parallels=10, coast_res='110m', cmap='RdBu_r'):
    """Display default quicklook plot.

    Parameters
    ---------
    area_def : object
        geometry.AreaDefinition object
    data : numpy array | numpy masked array
        2D array matching area_def. Use masked array for transparent values
    vmin : float, optional
        Min value for luminescence scaling
    vmax : float, optional
        Max value for luminescence scaling
    label : str, optional
        Label for data
    num_meridians : int, optional
        Number of meridians to plot on the globe
    num_parallels : int, optional
        Number of parallels to plot on the globe
    coast_res : {'c', 'l', 'i', 'h', 'f'}, optional
        Resolution of coastlines

    Returns
    -------
    bmap : Basemap object
    """
    plt = _get_quicklook(area_def, data, vmin=vmin, vmax=vmax,
                         label=label, num_meridians=num_meridians,
                         num_parallels=num_parallels, coast_res=coast_res,
                         cmap=cmap)
    plt.show()
    plt.close()


def save_quicklook(filename, area_def, data, vmin=None, vmax=None,
                   label='Variable (units)', num_meridians=45,
                   num_parallels=10, coast_res='110m',
                   cmap='RdBu_r'):
    """Display and save default quicklook plot.

    Parameters
    ----------
    filename : str
        path to output file
    area_def : object
        geometry.AreaDefinition object
    data : numpy array | numpy masked array
        2D array matching area_def. Use masked array for transparent values
    vmin : float, optional
        Min value for luminescence scaling
    vmax : float, optional
        Max value for luminescence scaling
    label : str, optional
        Label for data
    num_meridians : int, optional
        Number of meridians to plot on the globe
    num_parallels : int, optional
        Number of parallels to plot on the globe
    coast_res : {'c', 'l', 'i', 'h', 'f'}, optional
        Resolution of coastlines
    """
    plt = _get_quicklook(area_def, data, vmin=vmin, vmax=vmax,
                         label=label, num_meridians=num_meridians,
                         num_parallels=num_parallels, coast_res=coast_res,
                         cmap=cmap)
    plt.savefig(filename, bbox_inches='tight')
    plt.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Nearest neighbor resampler."""

from __future__ import annotations

import warnings
from copy import deepcopy
from logging import getLogger

import numpy as np
from pykdtree.kdtree import KDTree

from pyresample import CHUNK_SIZE, geometry
from pyresample.utils.errors import PerformanceWarning

from ..geometry import StaticGeometry, SwathDefinition
from ._transform_utils import lonlat2xyz
from .resampler import Resampler, update_resampled_coords

logger = getLogger(__name__)

try:
    import dask
    import dask.array as da
    from xarray import DataArray
except ImportError:
    DataArray = None
    da = None
    dask = None


def query_no_distance(target_lons, target_lats, valid_output_index,
                      mask=None, valid_input_index=None,
                      neighbours=None, epsilon=None, radius=None,
                      kdtree=None):
    """Query the kdtree. No distances are returned.

    NOTE: Dask array arguments must always come before other keyword arguments
          for `da.blockwise` arguments to work.
    """
    voi = valid_output_index
    shape = voi.shape + (neighbours,)
    voir = voi.ravel()
    if mask is not None:
        mask = mask.ravel()[valid_input_index.ravel()]
    target_lons_valid = target_lons.ravel()[voir]
    target_lats_valid = target_lats.ravel()[voir]

    coords = lonlat2xyz(target_lons_valid, target_lats_valid)
    distance_array, index_array = kdtree.query(
        coords,
        k=neighbours,
        eps=epsilon,
        distance_upper_bound=radius,
        mask=mask)

    if index_array.ndim == 1:
        index_array = index_array[:, None]

    # KDTree query returns out-of-bounds neighbors as `len(arr)`
    # which is an invalid index, we mask those out so -1 represents
    # invalid values
    # voi is 2D (trows, tcols)
    # index_array is 2D (valid output pixels, neighbors)
    # there are as many Trues in voi as rows in index_array
    good_pixels = index_array < kdtree.n
    res_ia = np.empty(shape, dtype=int)
    mask = np.zeros(shape, dtype=bool)
    mask[voi, :] = good_pixels
    res_ia[mask] = index_array[good_pixels]
    res_ia[~mask] = -1
    return res_ia


def _my_index(index_arr, vii, data_arr, vii_slices=None, ia_slices=None,
              fill_value=np.nan):
    """Wrap index logic for 'get_sample_from_neighbour_info' to be used inside dask map_blocks."""
    vii_slices = tuple(
        x if x is not None else vii.ravel() for x in vii_slices)
    mask_slices = tuple(
        x if x is not None else (index_arr == -1) for x in ia_slices)
    ia_slices = tuple(
        x if x is not None else index_arr for x in ia_slices)
    res = data_arr[vii_slices][ia_slices]
    res[mask_slices] = fill_value
    return res


# TODO: Add decorator for geom<->geom support
# TODO: Add decorator for object type support
# Must be decorators so that we can both add class attributes with this information
# and add decorators to __init__ and precompute/resample calls for validity checking
class KDTreeNearestXarrayResampler(Resampler):
    """Resampler using the basic nearest neighbor algorithm."""

    def __init__(self,
                 source_geo_def: StaticGeometry,
                 target_geo_def: StaticGeometry,
                 cache=None):
        """Resampler for xarray DataArrays using a nearest neighbor algorithm.

        Parameters
        ----------
        source_geo_def : object
            Geometry definition of source
        target_geo_def : object
            Geometry definition of target

        """
        if DataArray is None:
            raise ImportError("Missing 'xarray' and 'dask' dependencies")
        super().__init__(source_geo_def, target_geo_def, cache=cache)
        self._internal_cache = {}
        assert (self.target_geo_def.ndim == 2), \
            "Target area definition must be 2 dimensions"

    @property
    def version(self) -> str:
        """Get the current version of this class used for hashing and caching."""
        return "0.1"

    def _compute_radius_of_influence(self):
        """Estimate a good default radius_of_influence."""
        try:
            src_res = self.source_geo_def.geocentric_resolution()
        except RuntimeError:
            logger.warning("Could not calculate source definition resolution")
            src_res = np.nan
        try:
            dst_res = self.target_geo_def.geocentric_resolution()
        except RuntimeError:
            logger.warning("Could not calculate destination definition "
                           "resolution")
            dst_res = np.nan
        radius_of_influence = np.nanmax([src_res, dst_res])
        if np.isnan(radius_of_influence):
            logger.warning("Could not calculate radius_of_influence, falling "
                           "back to 10000 meters. This may produce lower "
                           "quality results than expected.")
            radius_of_influence = 10000
        return radius_of_influence

    def _create_resample_kdtree(self, chunks=CHUNK_SIZE):
        """Set up kd tree on input."""
        source_lons, source_lats = self.source_geo_def.get_lonlats(
            chunks=chunks)
        valid_input_idx = ((source_lons >= -180) & (source_lons <= 180) & (source_lats <= 90) & (source_lats >= -90))
        input_coords = lonlat2xyz(source_lons, source_lats)
        input_coords = input_coords[valid_input_idx.ravel(), :]

        # Build kd-tree on input
        input_coords = input_coords.astype(np.float64)
        delayed_kdtree = dask.delayed(KDTree, pure=True)(input_coords)
        return valid_input_idx, delayed_kdtree

    def _query_resample_kdtree(self,
                               resample_kdtree,
                               tlons,
                               tlats,
                               valid_input_index,
                               valid_output_index,
                               mask,
                               neighbors,
                               radius_of_influence,
                               epsilon):
        """Query kd-tree on slice of target coordinates."""
        if mask is None:
            args = tuple()
        else:
            ndims = self.source_geo_def.ndim
            dims = 'mn'[:ndims]
            args = (mask, dims, valid_input_index, dims)
        # res.shape = rows, cols, neighbors
        # j=rows, i=cols, k=neighbors, m=source rows, n=source cols
        res = da.blockwise(
            query_no_distance, 'jik', tlons, 'ji', tlats, 'ji',
            valid_output_index, 'ji', *args, kdtree=resample_kdtree,
            neighbours=neighbors, epsilon=epsilon,
            radius=radius_of_influence, dtype=np.int64,
            new_axes={'k': neighbors}, concatenate=True)
        return res

    def _get_neighbor_info(self, mask, neighbors, radius_of_influence, epsilon):
        """Return neighbour info.

        Returns
        -------
        (valid_input_index, valid_output_index,
        index_array, distance_array) : tuple of numpy arrays
            Neighbour resampling info
        """
        if self.source_geo_def.size < neighbors:
            warnings.warn('Searching for %s neighbors in %s data points' %
                          (neighbors, self.source_geo_def.size))

        # Create kd-tree
        chunks = mask.chunks if mask is not None else CHUNK_SIZE
        valid_input_idx, resample_kdtree = self._create_resample_kdtree(chunks=chunks)

        # TODO: Add 'chunks' keyword argument to this method and use it
        target_lons, target_lats = self.target_geo_def.get_lonlats(chunks=CHUNK_SIZE)
        valid_output_idx = ((target_lons >= -180) & (target_lons <= 180) & (target_lats <= 90) & (target_lats >= -90))

        if mask is not None:
            assert (mask.shape == self.source_geo_def.shape), \
                "'mask' must be the same shape as the source geo definition"
            mask = mask.data
        index_arr = self._query_resample_kdtree(
            resample_kdtree, target_lons, target_lats, valid_input_idx,
            valid_output_idx, mask,
            neighbors, radius_of_influence, epsilon)

        return valid_input_idx, index_arr

    def get_sample_from_neighbor_info(
            self,
            data,
            valid_input_index,
            index_array,
            neighbors=1,
            fill_value=np.nan):
        """Get the pixels matching the target area.

        This method should work for any dimensionality of the provided data
        array as long as the geolocation dimensions match in size and name in
        ``data.dims``. Where source area definition are `AreaDefinition`
        objects the corresponding dimensions in the data should be
        ``('y', 'x')``.

        This method also attempts to preserve chunk sizes of dask arrays,
        but does require loading/sharing the fully computed source data before
        it can actually compute the values to write to the destination array.
        This can result in large memory usage for large source data arrays,
        but is a necessary evil until fancier indexing is supported by dask
        and/or pykdtree.

        Args:
            data (xarray.DataArray): Source data pixels to sample
            fill_value (float): Output fill value when no source data is
                near the target pixel. When omitted, if the input data is an
                integer array then the maximum value for that integer type is
                used, but otherwise, NaN is used and can be detected in the
                result with ``res.isnull()``.

        Returns:
            dask.array.Array: The resampled array. The dtype of the array will
                be the same as the input data. Pixels with no matching data from
                the input array will be filled (see the `fill_value` parameter
                description above).
        """
        if fill_value is not None and np.isnan(fill_value) and \
                np.issubdtype(data.dtype, np.integer):
            fill_value = _get_fill_mask_value(data.dtype)
            logger.warning("Fill value incompatible with integer data "
                           "using {:d} instead.".format(fill_value))

        # Convert back to 1 neighbor
        if neighbors > 1:
            raise NotImplementedError("Nearest neighbor resampling can not "
                                      "handle more than 1 neighbor yet.")
        # Convert from multiple neighbor shape to 1 neighbor
        ia = index_array[:, :, 0]
        vii = valid_input_index

        src_geo_dims = self._get_src_geo_dims()
        dst_geo_dims = ('y', 'x')
        self._verify_data_geo_dims(data, src_geo_dims)

        # shape of the source data after we flatten the geo dimensions
        flat_src_shape = []
        # slice objects to index in to the source data
        vii_slices = []
        ia_slices = []
        # whether we have seen the geo dims in our analysis
        geo_handled = False
        # dimension indexes for da.blockwise
        src_adims = []
        flat_adim = []
        # map source dimension name to dimension number for da.blockwise
        src_dim_to_ind = {}
        # destination array dimension indexes for da.blockwise
        dst_dims = []
        for i, dim in enumerate(data.dims):
            src_dim_to_ind[dim] = i
            if dim in src_geo_dims and not geo_handled:
                flat_src_shape.append(-1)
                vii_slices.append(None)  # mark for replacement
                ia_slices.append(None)  # mark for replacement
                flat_adim.append(i)
                src_adims.append(i)
                dst_dims.extend(dst_geo_dims)
                geo_handled = True
            elif dim not in src_geo_dims:
                flat_src_shape.append(data.sizes[dim])
                vii_slices.append(slice(None))
                ia_slices.append(slice(None))
                src_adims.append(i)
                dst_dims.append(dim)
        # map destination dimension names to blockwise dimension indexes
        dst_dim_to_ind = src_dim_to_ind.copy()
        dst_dim_to_ind['y'] = i + 1
        dst_dim_to_ind['x'] = i + 2
        # FUTURE: when we allow more than one neighbor
        # neighbors_dim = i + 3

        new_data = data.data.reshape(flat_src_shape)
        vii = vii.ravel()
        dst_adims = [dst_dim_to_ind[dim] for dim in dst_dims]
        ia_adims = [dst_dim_to_ind[dim] for dim in dst_geo_dims]
        # FUTURE: when we allow more than one neighbor add neighbors dimension
        # dst_adims.append(neighbors_dim)
        # ia_adims.append(neighbors_dim)
        # FUTURE: when we allow more than one neighbor we need to add
        #         the new axis to blockwise:
        #         `new_axes={neighbor_dim: neighbors}`
        # FUTURE: if/when dask can handle index arrays that are dask arrays
        #         then we can avoid all of this complicated blockwise stuff
        res = da.blockwise(
            _my_index, dst_adims,
            ia, ia_adims,
            vii, flat_adim,
            new_data, src_adims,
            vii_slices=vii_slices, ia_slices=ia_slices,
            fill_value=fill_value,
            dtype=new_data.dtype, concatenate=True)
        res = DataArray(res, dims=dst_dims,
                        attrs=deepcopy(data.attrs))
        res = update_resampled_coords(data, res, self.target_geo_def)
        return res

    def _verify_data_geo_dims(self, data, src_geo_dims):
        # verify that source dims are the same between geo and data
        data_geo_dims = tuple(d for d in data.dims if d in src_geo_dims)
        if data_geo_dims != src_geo_dims:
            raise ValueError("Data dimensions do not match source area dimensions.")

        # verify that the dims are next to each other
        first_dim_idx = data.dims.index(src_geo_dims[0])
        num_dims = len(src_geo_dims)
        if data.dims[first_dim_idx:first_dim_idx + num_dims] != data_geo_dims:
            raise ValueError("Data's geolocation dimensions are not consecutive.")

        # verify that the dims are the same between src geom and data
        for dim_offset, dim_name in enumerate(data_geo_dims):
            geom_size = self.source_geo_def.shape[dim_offset]
            data_size = data.shape[first_dim_idx + dim_offset]
            if geom_size != data_size:
                raise ValueError("Input data shape is not equal to the shape of "
                                 "the source geometry: "
                                 f"{dim_name}={data_size} versus {geom_size}")

    def _get_src_geo_dims(self):
        if isinstance(self.source_geo_def, geometry.SwathDefinition):
            # could be 1D or 2D
            src_geo_dims = self.source_geo_def.lons.dims
        else:
            # assume AreaDefinitions and everything else are 2D with 'y', 'x'
            src_geo_dims = ('y', 'x')
        return src_geo_dims

    def precompute(self, mask=None, radius_of_influence=None, epsilon=0):
        """Generate neighbor indexes using geolocation information and optional data mask.

        Args:
            mask (ArrayLike, optional):
                Boolean array where True represents invalid pixels in the data
                array to be resampled in the future. This allows the indexes
                computed by this method and used during resampling to filter
                out invalid values and produce a result with more overall valid
                pixels. If provided then pre-computed results will not be
                cached as it is assumed that the mask will likely change for
                every input array.
            radius_of_influence (float, optional):
                Cut off distance in geocentric meters.
                If not provided this will be estimated based on the source
                and target geometry definition.
            epsilon (float, optional):
                Allowed uncertainty in meters. Increasing uncertainty
                reduces execution time

        """
        neighbors = 1
        if mask is not None and mask.shape != self.source_geo_def.shape:
            raise ValueError("'mask' provided to 'precompute' is not the same "
                             "shape as the source geometry.")
        if radius_of_influence is None:
            radius_of_influence = self._compute_radius_of_influence()

        # use dask task name
        mask_hash = None if mask is None else mask.data.name
        internal_cache_key = (mask_hash, neighbors, radius_of_influence, epsilon)
        in_int_cache = internal_cache_key in self._internal_cache
        if not in_int_cache:
            valid_input_index, index_arr = self._get_neighbor_info(
                mask, neighbors, radius_of_influence, epsilon)
            item_to_cache = {
                "valid_input_index": valid_input_index,
                "index_array": index_arr,
            }
            self._internal_cache[internal_cache_key] = item_to_cache

    def resample(self, data, mask_area=None, fill_value=np.nan,
                 radius_of_influence=None, epsilon=0):
        """Resample input ``data`` from the source geometry to the target geometry.

        Args:
            data (ArrayLike): Data to be resampled
            mask_area (bool or ArrayLike): Mask geolocation data where data
                values are invalid. This should be used when data values may
                affect what neighbors are considered valid. For complex masking
                behavior this can also be the mask array to use instead of the
                resampler computing it on the fly. This is useful
                for non-xarray DataArrays that don't have related
                metadata like "_FillValue". By default this is None which means
                a mask will be created automatically for SwathDefinition input.
                Set to ``False`` to disable any masking and ``True`` to ensure
                a mask is created. See :meth:`precompute` for more information.
            fill_value (int or float): Output fill value when no source data is
                near the target pixel. When omitted, if the input data is an
                integer array then the maximum value for that integer type is
                used, but otherwise, NaN is used and can be detected in the
                result with ``res.isnull()`` for DataArray output and
                ``np.isnan(res)`` for dask and numpy arrays.
            radius_of_influence (float, optional):
                Passed directly to :meth:`precompute`.
            epsilon (float, optional):
                Passed directly to :meth:`precompute`.

        Returns:
            Array-like object of the same type as ``data``, resampled to the
            target geographic geometry.

        """
        new_data = self._verify_input_object_type(data)
        src_geo_dims = self._get_src_geo_dims()
        self._verify_data_geo_dims(new_data, src_geo_dims)

        mask = self._get_area_mask(mask_area, new_data)
        if radius_of_influence is None:
            radius_of_influence = self._compute_radius_of_influence()
        self.precompute(mask=mask, radius_of_influence=radius_of_influence, epsilon=epsilon)

        # Get precomputed arrays - use dask array task name
        mask_hash = None if mask is None else mask.data.name
        cache_key = (mask_hash, 1, radius_of_influence, epsilon)
        precompute_dict = self._internal_cache[cache_key]
        result = self.get_sample_from_neighbor_info(
            new_data,
            precompute_dict["valid_input_index"],
            precompute_dict["index_array"],
            fill_value=fill_value)
        return self._verify_result_object_type(result, data)

    def _verify_input_object_type(self, data):
        if isinstance(data, DataArray) and isinstance(data.data, da.Array):
            return data
        if not isinstance(data, DataArray):
            if data.ndim != 2:
                raise ValueError(
                    f"{self.__class__.__name__} requires DataArrays with a dask "
                    "array. Input array is not 2D and can't be automatically "
                    "converted.")
            data = DataArray(data, dims=self._get_src_geo_dims())
        if not isinstance(data.data, da.Array):
            warnings.warn(
                f"{self.__class__.__name__} uses a dask-based implementation, "
                "but a pure numpy array was provided. Data will be converted "
                "to dask arrays for computation and then converted back. To "
                "avoid this warning convert your numpy array before providing "
                "it to the resampler.", PerformanceWarning)
            data = data.copy()
            data.data = da.from_array(data.data, chunks="auto")
        return data

    def _verify_result_object_type(self, data, orig_data):
        if isinstance(orig_data, DataArray) and isinstance(orig_data.data, da.Array):
            return data
        if not isinstance(orig_data, DataArray):
            data = data.data
        if isinstance(orig_data, da.Array):
            return data
        return data.compute()

    def _get_area_mask(self, mask_area, data):
        if isinstance(mask_area, (np.ndarray, da.Array, DataArray)):
            return mask_area

        # default is to mask areas for SwathDefinitions
        if mask_area is None and isinstance(self.source_geo_def, SwathDefinition):
            mask_area = True

        if mask_area:
            return self.compute_data_mask(data)

    def compute_data_mask(self, data):
        """Generate a mask array where data is invalid.

        This is used by :meth:`resample` to determine what ``mask`` to pass
        to :meth:`precompute`. It may be useful for users to use this manually
        for special cases of wanting to call `precompute` manually.

        """
        geo_dims = self._get_src_geo_dims()
        flat_dims = [dim for dim in data.dims if dim not in geo_dims]
        if np.issubdtype(data.dtype, np.integer):
            mask = data == data.attrs.get('_FillValue', np.iinfo(data.dtype.type).max)
        else:
            mask = data.isnull()
        mask = mask.all(dim=flat_dims)
        return mask


def _get_fill_mask_value(data_dtype):
    """Return the maximum value of dtype."""
    if issubclass(data_dtype.type, np.floating):
        fill_value = np.finfo(data_dtype.type).max
    elif issubclass(data_dtype.type, np.integer):
        fill_value = np.iinfo(data_dtype.type).max
    else:
        raise TypeError('Type %s is unsupported for masked fill values' %
                        data_dtype.type)
    return fill_value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pyresample, Resampling of remote sensing image data in python
#
# Copyright (C) 2014-2019 PyTroll developers
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Testing the _spatial_mp module."""

import unittest

import numpy as np

import pyresample._spatial_mp as sp


class SpatialMPTest(unittest.TestCase):
    """Test of spatial_mp."""

    def setUp(self):
        """Set up some variables."""
        self.exp_coords = np.array([[6370997., 0, 0],
                                    [6178887.9339746, 1089504.6535337, 1106312.0189715],
                                    [5233097.4664751, 2440233.4244888, 2692499.6776952]])
        self.lon = np.array([0, 10, 25])
        self.lat = np.array([0, 10, 25])
        self.lon32 = self.lon.astype(np.float32)
        self.lat32 = self.lon.astype(np.float32)
        self.lon64 = self.lon.astype(np.float64)
        self.lat64 = self.lon.astype(np.float64)

    def test_cartesian(self):
        """Test the transform_lonlats of class Cartesian."""
        my_cartesian = sp.Cartesian()
        coords_int = my_cartesian.transform_lonlats(self.lon, self.lat)
        coords_float = my_cartesian.transform_lonlats(self.lon64, self.lat64)
        coords_float32 = my_cartesian.transform_lonlats(self.lon32, self.lat32)
        np.testing.assert_almost_equal(coords_float, self.exp_coords, decimal=3)
        np.testing.assert_almost_equal(coords_int, coords_float, decimal=0)
        self.assertIs(type(coords_float32[0, 0]), np.float32)
        self.assertIs(type(coords_float[0, 0]), np.float64)
        self.assertTrue(np.issubdtype(coords_int.dtype, np.floating))

    def test_cartesian_without_ne(self):
        """Test the transform_lonlats of class Cartesian without numexpr."""
        sp.ne = False
        my_cartesian = sp.Cartesian()
        coords_int = my_cartesian.transform_lonlats(self.lon, self.lat)
        coords_float = my_cartesian.transform_lonlats(self.lon64, self.lat64)
        coords_float32 = my_cartesian.transform_lonlats(self.lon32, self.lat32)
        np.testing.assert_almost_equal(coords_float, self.exp_coords, decimal=3)
        np.testing.assert_almost_equal(coords_int, coords_float, decimal=0)
        self.assertIs(type(coords_float32[0, 0]), np.float32)
        self.assertIs(type(coords_float[0, 0]), np.float64)
        self.assertTrue(np.issubdtype(coords_int.dtype, np.floating))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016

# Author(s):

#   David Hoese <david.hoese@ssec.wisc.edu>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Map longitude/latitude points to column/rows of a grid."""
__docformat__ = "restructuredtext en"

import numpy
from pyproj import Proj

cimport cython
cimport numpy

# column and rows can only be doubles for now until the PROJ.4 is linked directly so float->double casting can be done
# inside the loop
ctypedef fused cr_dtype:
    # numpy.float32_t
    numpy.float64_t

cdef extern from "numpy/npy_math.h":
    bint npy_isnan(double x)


def projection_circumference(p):
    """Return the projection circumference if the projection is cylindrical, None otherwise.

    Projections that are not cylindrical and centered on the globes axis
    can not easily have data cross the antimeridian of the projection.
    """
    lon0, lat0 = p(0, 0, inverse=True)
    lon1 = lon0 + 180.0
    lat1 = lat0 + 5.0
    x0, y0 = p(lon0, lat0)  # should result in zero or near zero
    x1, y1 = p(lon1, lat0)
    x2, y2 = p(lon1, lat1)
    if y0 != y1 or x1 != x2:
        return 0.0
    return abs(x1 - x0) * 2


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def ll2cr_dynamic(numpy.ndarray[cr_dtype, ndim=2] lon_arr, numpy.ndarray[cr_dtype, ndim=2] lat_arr,
                  cr_dtype fill_in, str proj4_definition,
                  double cell_width, double cell_height,
                  width=None, height=None,
                  origin_x=None, origin_y=None):
    """Project longitude and latitude points to column rows in the specified grid in place.

    This function is meant to operate on dynamic grids and is theoretically
    slower than the `ll2cr_static` function. Dynamic grids are those that
    have one or more set of parameters unspecified like the number of pixels
    in the grid (width, height) or the origin coordinates of the grid
    (origin_x, origin_y). This function will analyze the provided lon/lat data
    and determine what the missing parameters must be so all input swath data
    falls in the resulting grid.

    :param lon_arr: Numpy array of longitude floats
    :param lat_arr: Numpy array of latitude floats
    :param grid_info: dictionary of grid information (see below)
    :param fill_in: Fill value for input longitude and latitude arrays and used for output
    :param proj4_definition: PROJ.4 string projection definition
    :param cell_width: Pixel resolution in the X direction in projection space
    :param cell_height: Pixel resolution in the Y direction in projection space
    :param width: (optional) Number of pixels in the X direction in the final output grid
    :param height: (optional) Number of pixels in the Y direction in the final output grid
    :param origin_x: (optional) Grid X coordinate for the upper-left pixel of the output grid
    :param origin_y: (optional) Grid Y coordinate for the upper-left pixel of the output grid
    :returns: tuple(points_in_grid, cols_out, rows_out, origin_x, origin_y, width, height)

    Steps taken in this function:

        1. Convert (lon, lat) points to (X, Y) points in the projection space
        2. If grid is missing some parameters (dynamic grid), then fill them in
        3. Convert (X, Y) points to (column, row) points in the grid space

    Note longitude and latitude arrays are limited to 64-bit floats because
    of limitations in pyproj.
    """
    # pure python stuff for now
    p = Proj(proj4_definition)

    # Pyproj currently makes a copy so we don't have to do anything special here
    cdef tuple projected_tuple = p(lon_arr, lat_arr)
    cdef cr_dtype[:, ::1] rows_out = projected_tuple[1]
    cdef cr_dtype[:, ::1] cols_out = projected_tuple[0]
    cdef double proj_circum = projection_circumference(p)
    cdef unsigned int w
    cdef unsigned int h
    cdef double ox
    cdef double oy

    # indexes
    cdef unsigned int row
    cdef unsigned int col
    # index bounds
    cdef unsigned int num_rows = lon_arr.shape[0]
    cdef unsigned int num_cols = lon_arr.shape[1]
    cdef cr_dtype xmin = cols_out[0, 0]
    cdef cr_dtype xmax = cols_out[0, 0]
    cdef cr_dtype ymin = rows_out[0, 0]
    cdef cr_dtype ymax = rows_out[0, 0]
    cdef cr_dtype x_tmp
    cdef cr_dtype y_tmp
    cdef unsigned int points_in_grid = 0
    for row in range(num_rows):
        for col in range(num_cols):
            x_tmp = cols_out[row, col]
            y_tmp = rows_out[row, col]

            if x_tmp >= 1e30:
                # pyproj library should have set both x and y to the fill value
                # we technically don't ever check for the fill value, but if fill values are valid lon/lats then WTF
                continue
            elif x_tmp < xmin or npy_isnan(xmin):
                xmin = x_tmp
            elif x_tmp > xmax or npy_isnan(xmax) or xmax >= 1e30:
                # Note: technically 2 valid points are required to get here if there are a lot of NaNs
                xmax = x_tmp

            if y_tmp < ymin or npy_isnan(ymin):
                ymin = y_tmp
            elif y_tmp > ymax or npy_isnan(ymax) or ymax >= 1e30:
                # Note: technically 2 valid points are required to get here if there are a lot of NaNs
                ymax = y_tmp

    # Check if we cross the antimeridian
    if proj_circum != 0 and xmax - xmin >= proj_circum * .75:
        # xmax will increase, but we need to reset xmin so that it gets properly detected
        if xmin < 0:
            xmin = xmax
        for row in range(num_rows):
            for col in range(num_cols):
                x_tmp = cols_out[row, col]
                if x_tmp < 0:
                    x_tmp += proj_circum
                    cols_out[row, col] = x_tmp
                    # xmax won't increase unless we've added the circumference
                    if x_tmp > xmax:
                        xmax = x_tmp
                elif x_tmp >= 1e30:
                    continue
                elif x_tmp < xmin:
                    # xmin could change with any of the remaining entries
                    xmin = x_tmp

    if origin_x is None:
        # upper-left corner
        ox = xmin
        oy = ymax
    else:
        ox = origin_x
        oy = origin_y

    if width is None:
        w = int(abs((xmax - ox) / cell_width))
        h = int(abs((oy - ymin) / cell_height))
    else:
        w = width
        h = height

    for row in range(num_rows):
        for col in range(num_cols):
            x_tmp = cols_out[row, col]
            y_tmp = rows_out[row, col]
            if x_tmp >= 1e30:
                lon_arr[row, col] = fill_in
                lat_arr[row, col] = fill_in
                continue

            x_tmp = (x_tmp - ox) / cell_width
            y_tmp = (y_tmp - oy) / cell_height
            if x_tmp >= -1 and x_tmp <= w + 1 and y_tmp >= -1 and y_tmp <= h + 1:
                points_in_grid += 1
            lon_arr[row, col] = x_tmp
            lat_arr[row, col] = y_tmp

    # return points_in_grid, x_arr, y_arr
    return points_in_grid, lon_arr, lat_arr, ox, oy, w, h


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def ll2cr_static(numpy.ndarray[cr_dtype, ndim=2] lon_arr, numpy.ndarray[cr_dtype, ndim=2] lat_arr,
                 cr_dtype fill_in, str proj4_definition,
                 double cell_width, double cell_height,
                 unsigned int width, unsigned int height,
                 double origin_x, double origin_y):
    """Project longitude and latitude points to column rows in the specified grid in place.

    :param lon_arr: Numpy array of longitude floats
    :param lat_arr: Numpy array of latitude floats
    :param fill_in: Fill value for input longitude and latitude arrays and used for output
    :param cell_width: Pixel resolution in the X direction in projection space
    :param cell_height: Pixel resolution in the Y direction in projection space
    :param width: Number of pixels in the X direction in the final output grid
    :param height: Number of pixels in the Y direction in the final output grid
    :param origin_x: Grid X coordinate for the upper-left pixel of the output grid
    :param origin_y: Grid Y coordinate for the upper-left pixel of the output grid
    :returns: tuple(points_in_grid, cols_out, rows_out)

    Steps taken in this function:

        1. Convert (lon, lat) points to (X, Y) points in the projection space
        2. Convert (X, Y) points to (column, row) points in the grid space

    Note longitude and latitude arrays are limited to 64-bit floats because
    of limitations in pyproj.
    """
    # pure python stuff for now
    p = Proj(proj4_definition)

    # Pyproj currently makes a copy so we don't have to do anything special here
    cdef tuple projected_tuple = p(lon_arr, lat_arr)
    cdef cr_dtype[:, ::1] rows_out = projected_tuple[1]
    cdef cr_dtype[:, ::1] cols_out = projected_tuple[0]

    # indexes
    cdef unsigned int row
    cdef unsigned int col
    # index bounds
    cdef unsigned int num_rows = lon_arr.shape[0]
    cdef unsigned int num_cols = lon_arr.shape[1]
    cdef cr_dtype x_tmp
    cdef cr_dtype y_tmp
    cdef unsigned int points_in_grid = 0

    for row in range(num_rows):
        for col in range(num_cols):
            x_tmp = cols_out[row, col]
            y_tmp = rows_out[row, col]
            if x_tmp >= 1e30:
                lon_arr[row, col] = fill_in
                lat_arr[row, col] = fill_in
                continue

            x_tmp = (x_tmp - origin_x) / cell_width
            y_tmp = (y_tmp - origin_y) / cell_height
            if x_tmp >= -1 and x_tmp <= width + 1 and y_tmp >= -1 and y_tmp <= height + 1:
                points_in_grid += 1
            lon_arr[row, col] = x_tmp
            lat_arr[row, col] = y_tmp

    # return points_in_grid, x_arr, y_arr
    return points_in_grid



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2010-2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Helper for working with chunked data over multiple workers."""

from __future__ import absolute_import

import ctypes
import multiprocessing as mp

import numpy as np


class Scheduler(object):
    """Provide chunks of data to work on based on various 'schedule' algorithms."""

    def __init__(self, ndata, nprocs, chunk=None, schedule='guided'):
        if schedule not in ['guided', 'dynamic', 'static']:
            raise ValueError('unknown scheduling strategy')
        self._ndata = mp.RawValue(ctypes.c_int, ndata)
        self._start = mp.RawValue(ctypes.c_int, 0)
        self._lock = mp.Lock()
        self._schedule = schedule
        self._nprocs = nprocs
        if schedule == 'guided' or schedule == 'dynamic':
            min_chunk = ndata // (10 * nprocs)
            if chunk:
                min_chunk = chunk
            min_chunk = max(min_chunk, 1)
            self._chunk = min_chunk
        elif schedule == 'static':
            min_chunk = ndata // nprocs
            if chunk:
                min_chunk = max(chunk, min_chunk)
            min_chunk = max(min_chunk, 1)
            self._chunk = min_chunk

    def __iter__(self):
        """Iterate over individual chunks of data."""
        while True:
            self._lock.acquire()
            ndata = self._ndata.value
            nprocs = self._nprocs
            start = self._start.value
            if self._schedule == 'guided':
                _chunk = ndata // nprocs
                chunk = max(self._chunk, _chunk)
            else:
                chunk = self._chunk
            if ndata:
                if chunk > ndata:
                    s0 = start
                    s1 = start + ndata
                    self._ndata.value = 0
                else:
                    s0 = start
                    s1 = start + chunk
                    self._ndata.value = ndata - chunk
                    self._start.value = start + chunk
                self._lock.release()
                yield slice(s0, s1)
            else:
                self._lock.release()
                return


def shmem_as_ndarray(raw_array):
    """Create numpy array from shared memory data."""
    _ctypes_to_numpy = {
        ctypes.c_char: np.int8,
        ctypes.c_wchar: np.int16,
        ctypes.c_byte: np.int8,
        ctypes.c_ubyte: np.uint8,
        ctypes.c_short: np.int16,
        ctypes.c_ushort: np.uint16,
        ctypes.c_int: np.int32,
        ctypes.c_uint: np.int32,
        ctypes.c_long: np.int32,
        ctypes.c_ulong: np.int32,
        ctypes.c_float: np.float32,
        ctypes.c_double: np.float64
    }
    dtype = _ctypes_to_numpy[raw_array._type_]

    # The following works too, but occasionally raises
    # RuntimeWarning: Item size computed from the PEP 3118 buffer format string does not match the actual item size.
    # and appears to be slower.
    # return np.ctypeslib.as_array(raw_array)

    return np.frombuffer(raw_array, dtype=dtype)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2010-2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Pyresample package for geographic data resampling and related utilities."""

import os

CHUNK_SIZE = int(os.getenv('PYTROLL_CHUNK_SIZE', 4096))

# Backwards compatibility
from pyresample import geometry  # noqa
from pyresample import grid  # noqa
from pyresample import image  # noqa
from pyresample import kd_tree  # noqa
from pyresample import plot  # noqa
from pyresample import utils  # noqa
from pyresample.area_config import (  # noqa
    convert_def_to_yaml,
    create_area_def,
    get_area_def,
    load_area,
    parse_area_file,
)

# Easy access
from pyresample.geometry import AreaDefinition  # noqa
from pyresample.geometry import DynamicAreaDefinition  # noqa
from pyresample.geometry import SwathDefinition  # noqa
from pyresample.kd_tree import XArrayResamplerNN  # noqa
from pyresample.plot import area_def2basemap, save_quicklook  # noqa

from .version import get_versions  # noqa

__all__ = ['grid', 'image', 'kd_tree', 'utils', 'plot', 'geo_filter', 'geometry', 'CHUNK_SIZE',
           'load_area', 'create_area_def', 'get_area_def', 'parse_area_file', 'convert_def_to_yaml']

__version__ = get_versions()['version']
del get_versions



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2013-2021 Pyresample Developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test cases for spherical geometry."""

import unittest

import numpy as np

from pyresample.spherical import Arc, CCoordinate, SCoordinate, SphPolygon

VERTICES_TEST_IS_INSIDE1 = np.array([[-1.54009253, 82.62402855],
                                     [3.4804808, 82.8105746],
                                     [20.7214892, 83.00875812],
                                     [32.8857629, 82.7607758],
                                     [41.53844302, 82.36024339],
                                     [47.92062759, 81.91317164],
                                     [52.82785062, 81.45769791],
                                     [56.75107895, 81.00613046],
                                     [59.99843787, 80.56042986],
                                     [62.76998034, 80.11814453],
                                     [65.20076209, 79.67471372],
                                     [67.38577498, 79.22428],
                                     [69.39480149, 78.75981318],
                                     [71.28163984, 78.27283234],
                                     [73.09016378, 77.75277976],
                                     [74.85864685, 77.18594725],
                                     [76.62327682, 76.55367303],
                                     [78.42162204, 75.82918893],
                                     [80.29698409, 74.97171721],
                                     [82.30538638, 73.9143231],
                                     [84.52973107, 72.53535661],
                                     [87.11696138, 70.57600156],
                                     [87.79163209, 69.98712409],
                                     [72.98142447, 67.1760143],
                                     [61.79517279, 63.2846272],
                                     [53.50600609, 58.7098766],
                                     [47.26725347, 53.70533139],
                                     [42.44083259, 48.42199571],
                                     [38.59682041, 42.95008531],
                                     [35.45189206, 37.3452509],
                                     [32.43435578, 30.72373327],
                                     [31.73750748, 30.89485287],
                                     [29.37284023, 31.44344415],
                                     [27.66001308, 31.81016309],
                                     [26.31358296, 32.08057499],
                                     [25.1963477, 32.29313986],
                                     [24.23118049, 32.46821821],
                                     [23.36993508, 32.61780082],
                                     [22.57998837, 32.74952569],
                                     [21.8375532, 32.86857867],
                                     [21.12396693, 32.97868717],
                                     [20.42339605, 33.08268331],
                                     [19.72121983, 33.18284728],
                                     [19.00268283, 33.28113306],
                                     [18.2515215, 33.3793305],
                                     [17.4482606, 33.47919405],
                                     [16.56773514, 33.58255576],
                                     [15.57501961, 33.6914282],
                                     [14.4180087, 33.8080799],
                                     [13.01234319, 33.93498577],
                                     [11.20625437, 34.0742239],
                                     [8.67990371, 34.22415978],
                                     [7.89344478, 34.26018768],
                                     [8.69446485, 41.19823568],
                                     [9.25707165, 47.17351118],
                                     [9.66283477, 53.14128114],
                                     [9.84134875, 59.09937166],
                                     [9.65054241, 65.04458004],
                                     [8.7667375, 70.97023122],
                                     [6.28280904, 76.85731403]], dtype='float64')

VERTICES_TEST_IS_INSIDE2 = np.array([[49.94506701, 46.52610743],
                                     [51.04293649, 46.52610743],
                                     [62.02163129, 46.52610743],
                                     [73.0003261, 46.52610743],
                                     [83.9790209, 46.52610743],
                                     [85.05493299, 46.52610743],
                                     [85.05493299, 45.76549301],
                                     [85.05493299, 37.58315571],
                                     [85.05493299, 28.39260587],
                                     [85.05493299, 18.33178739],
                                     [85.05493299, 17.30750918],
                                     [83.95706351, 17.30750918],
                                     [72.97836871, 17.30750918],
                                     [61.9996739, 17.30750918],
                                     [51.0209791, 17.30750918],
                                     [49.94506701, 17.30750918],
                                     [49.94506701, 18.35262921],
                                     [49.94506701, 28.41192025],
                                     [49.94506701, 37.60055422],
                                     [49.94506701, 45.78080831]], dtype='float64')


class TestSCoordinate(unittest.TestCase):
    """Test SCoordinates."""

    def test_distance(self):
        """Test Vincenty formula."""
        d = SCoordinate(0, 0).distance(SCoordinate(1, 1))
        np.testing.assert_equal(d, 1.2745557823062943)

    def test_hdistance(self):
        """Test Haversine formula."""
        d = SCoordinate(0, 0).hdistance(SCoordinate(1, 1))
        self.assertTrue(np.allclose(d, 1.2745557823062943))

    def test_str(self):
        """Check the string representation."""
        d = SCoordinate(0, 0)
        self.assertEqual(str(d), "(0.0, 0.0)")

    def test_repr(self):
        """Check the representation."""
        d = SCoordinate(0, 0)
        self.assertEqual(repr(d), "(0.0, 0.0)")


class TestCCoordinate(unittest.TestCase):
    """Test SCoordinates."""

    def test_str(self):
        """Check the string representation."""
        d = CCoordinate((0, 0, 0))
        self.assertEqual(str(d), "[0 0 0]")

    def test_repr(self):
        """Check the representation."""
        d = CCoordinate((0, 0, 0))
        self.assertEqual(repr(d), "[0 0 0]")

    def test_norm(self):
        """Euclidean norm of a cartesian vector."""
        d = CCoordinate((1, 0, 0))
        self.assertEqual(d.norm(), 1.0)

    def test_normalize(self):
        """Normalize a cartesian vector."""
        d = CCoordinate((2., 0., 0.))
        self.assertTrue(np.allclose(d.normalize().cart, [1, 0, 0]))

    def test_cross(self):
        """Test cross product in cartesian coordinates."""
        d = CCoordinate((1., 0., 0.))
        c = CCoordinate((0., 1., 0.))
        self.assertTrue(np.allclose(d.cross(c).cart, [0., 0., 1.]))

    def test_dot(self):
        """Test the dot product of two cartesian vectors."""
        d = CCoordinate((1., 0., 0.))
        c = CCoordinate((0., 1., 0.))
        self.assertEqual(d.dot(c), 0)

    def test_ne(self):
        """Test inequality of two cartesian vectors."""
        d = CCoordinate((1., 0., 0.))
        c = CCoordinate((0., 1., 0.))
        self.assertTrue(c != d)

    def test_eq(self):
        """Test equality of two cartesian vectors."""
        d = CCoordinate((1., 0., 0.))
        c = CCoordinate((0., 1., 0.))
        self.assertFalse(c == d)

    def test_add(self):
        """Test adding cartesian vectors."""
        d = CCoordinate((1., 0., 0.))
        c = CCoordinate((0., 1., 0.))
        b = CCoordinate((1., 1., 0.))
        self.assertTrue(np.allclose((d + c).cart, b.cart))

        self.assertTrue(np.allclose((d + (0, 1, 0)).cart, b.cart))

        self.assertTrue(np.allclose(((0, 1, 0) + d).cart, b.cart))

    def test_mul(self):
        """Test multiplying (element-wise) cartesian vectors."""
        d = CCoordinate((1., 0., 0.))
        c = CCoordinate((0., 1., 0.))
        b = CCoordinate((0., 0., 0.))
        self.assertTrue(np.allclose((d * c).cart, b.cart))
        self.assertTrue(np.allclose((d * (0, 1, 0)).cart, b.cart))

        self.assertTrue(np.allclose(((0, 1, 0) * d).cart, b.cart))

    def test_to_spherical(self):
        """Test converting to spherical coordinates."""
        d = CCoordinate((1., 0., 0.))
        c = SCoordinate(0, 0)
        self.assertEqual(d.to_spherical(), c)


class TestArc(unittest.TestCase):
    """Test arcs."""

    def test_eq(self):
        arc1 = Arc(SCoordinate(0, 0),
                   SCoordinate(np.deg2rad(10), np.deg2rad(10)))
        arc2 = Arc(SCoordinate(0, np.deg2rad(10)),
                   SCoordinate(np.deg2rad(10), 0))

        self.assertFalse(arc1 == arc2)

        self.assertTrue(arc1 == arc1)

    def test_ne(self):
        arc1 = Arc(SCoordinate(0, 0),
                   SCoordinate(np.deg2rad(10), np.deg2rad(10)))
        arc2 = Arc(SCoordinate(0, np.deg2rad(10)),
                   SCoordinate(np.deg2rad(10), 0))

        self.assertTrue(arc1 != arc2)

        self.assertFalse(arc1 != arc1)

    def test_str(self):
        arc1 = Arc(SCoordinate(0, 0),
                   SCoordinate(np.deg2rad(10), np.deg2rad(10)))
        self.assertEqual(str(arc1), str(arc1.start) + " -> " + str(arc1.end))
        self.assertEqual(repr(arc1), str(arc1.start) + " -> " + str(arc1.end))

    def test_intersection(self):
        arc1 = Arc(SCoordinate(0, 0),
                   SCoordinate(np.deg2rad(10), np.deg2rad(10)))
        arc2 = Arc(SCoordinate(0, np.deg2rad(10)),
                   SCoordinate(np.deg2rad(10), 0))
        lon, lat = arc1.intersection(arc2)

        self.assertTrue(np.allclose(np.rad2deg(lon), 5))
        self.assertEqual(np.rad2deg(lat).round(7), round(5.0575148968282093, 7))

        arc1 = Arc(SCoordinate(0, 0),
                   SCoordinate(np.deg2rad(10), np.deg2rad(10)))

        self.assertTrue(arc1.intersection(arc1) is None)

        arc1 = Arc(SCoordinate(np.deg2rad(24.341215776575297),
                               np.deg2rad(44.987819588259327)),
                   SCoordinate(np.deg2rad(18.842727517611817),
                               np.deg2rad(46.512483610284178)))
        arc2 = Arc(SCoordinate(np.deg2rad(20.165961750361905),
                               np.deg2rad(46.177305385810541)),
                   SCoordinate(np.deg2rad(20.253297585831707),
                               np.deg2rad(50.935830837274324)))
        inter = SCoordinate(np.deg2rad(20.165957021925202),
                            np.deg2rad(46.177022633103398))
        self.assertEqual(arc1.intersection(arc2), inter)

        arc1 = Arc(SCoordinate(np.deg2rad(-2.4982818108326734),
                               np.deg2rad(48.596644847869655)),
                   SCoordinate(np.deg2rad(-2.9571441235622835),
                               np.deg2rad(49.165688435261394)))
        arc2 = Arc(SCoordinate(np.deg2rad(-3.4976667413531688),
                               np.deg2rad(48.562704872921373)),
                   SCoordinate(np.deg2rad(-5.893976312685715),
                               np.deg2rad(48.445795283217116)))

        self.assertTrue(arc1.intersection(arc2) is None)

    def test_angle(self):
        arc1 = Arc(SCoordinate(np.deg2rad(157.5),
                               np.deg2rad(89.234600944314138)),
                   SCoordinate(np.deg2rad(90),
                               np.deg2rad(89)))
        arc2 = Arc(SCoordinate(np.deg2rad(157.5),
                               np.deg2rad(89.234600944314138)),
                   SCoordinate(np.deg2rad(135),
                               np.deg2rad(89)))

        self.assertAlmostEqual(np.rad2deg(arc1.angle(arc2)), -44.996385007218926)

        arc1 = Arc(SCoordinate(np.deg2rad(112.5),
                               np.deg2rad(89.234600944314138)),
                   SCoordinate(np.deg2rad(90), np.deg2rad(89)))
        arc2 = Arc(SCoordinate(np.deg2rad(112.5),
                               np.deg2rad(89.234600944314138)),
                   SCoordinate(np.deg2rad(45), np.deg2rad(89)))

        self.assertAlmostEqual(np.rad2deg(arc1.angle(arc2)), 44.996385007218883)

        arc1 = Arc(SCoordinate(0, 0), SCoordinate(1, 0))
        self.assertAlmostEqual(arc1.angle(arc1), 0)

        arc2 = Arc(SCoordinate(1, 0), SCoordinate(0, 0))
        self.assertAlmostEqual(arc1.angle(arc2), 0)

        arc2 = Arc(SCoordinate(0, 0), SCoordinate(-1, 0))
        self.assertAlmostEqual(arc1.angle(arc2), np.pi)

        arc2 = Arc(SCoordinate(2, 0), SCoordinate(1, 0))
        self.assertAlmostEqual(arc1.angle(arc2), np.pi)

        arc2 = Arc(SCoordinate(2, 0), SCoordinate(3, 0))
        self.assertRaises(ValueError, arc1.angle, arc2)


class TestSphericalPolygon(unittest.TestCase):
    """Test the spherical polygon."""

    def test_area(self):
        """Test the area function."""
        vertices = np.array([[1, 2, 3, 4, 3, 2],
                             [3, 4, 3, 2, 1, 2]]).T
        polygon = SphPolygon(np.deg2rad(vertices))

        self.assertAlmostEqual(0.00121732523118, polygon.area())

        vertices = np.array([[1, 2, 3, 2],
                             [3, 4, 3, 2]]).T
        polygon = SphPolygon(np.deg2rad(vertices))

        self.assertAlmostEqual(0.000608430665842, polygon.area())

        vertices = np.array([[0, 0, 1, 1],
                             [0, 1, 1, 0]]).T
        polygon = SphPolygon(np.deg2rad(vertices))

        self.assertAlmostEqual(0.000304609684862, polygon.area())

        # Across the dateline

        vertices = np.array([[179.5, -179.5, -179.5, 179.5],
                             [1, 1, 0, 0]]).T
        polygon = SphPolygon(np.deg2rad(vertices))

        self.assertAlmostEqual(0.000304609684862, polygon.area())

        vertices = np.array([[0, 90, 90, 0],
                             [1, 1, 0, 0]]).T
        polygon = SphPolygon(np.deg2rad(vertices))

        self.assertAlmostEqual(0.0349012696772, polygon.area())

        vertices = np.array([[90, 0, 0],
                             [0, 0, 90]]).T
        polygon = SphPolygon(np.deg2rad(vertices))

        self.assertAlmostEqual(np.pi / 2, polygon.area())

        # Around the north pole

        vertices = np.array([[0, -90, 180, 90],
                             [89, 89, 89, 89]]).T
        polygon = SphPolygon(np.deg2rad(vertices))

        self.assertAlmostEqual(0.000609265770322, polygon.area())

        # Around the south pole

        vertices = np.array([[0, 90, 180, -90],
                             [-89, -89, -89, -89]]).T
        polygon = SphPolygon(np.deg2rad(vertices))

        self.assertAlmostEqual(0.000609265770322, polygon.area())

    def test_is_inside(self):
        """Test checking if a polygon is inside of another."""
        vertices = np.array([[1, 1, 20, 20],
                             [1, 20, 20, 1]]).T

        rad_verts = np.deg2rad(vertices)
        polygon1 = SphPolygon(rad_verts)
        # make sure 64-bit floats don't get copied
        assert polygon1.vertices is rad_verts

        vertices = np.array([[0, 0, 30, 30],
                             [0, 30, 30, 0]]).T

        polygon2 = SphPolygon(np.deg2rad(vertices))

        self.assertTrue(polygon1._is_inside(polygon2))

        self.assertFalse(polygon2._is_inside(polygon1))

        # Why checking the areas here!? It has nothing to do with the is_inside function!
        self.assertTrue(polygon2.area() > polygon1.area())

        polygon2.invert()
        self.assertFalse(polygon1._is_inside(polygon2))
        self.assertFalse(polygon2._is_inside(polygon1))

        vertices = np.array([[0, 0, 30, 30],
                             [21, 30, 30, 21]]).T

        polygon2 = SphPolygon(np.deg2rad(vertices))
        self.assertFalse(polygon1._is_inside(polygon2))
        self.assertFalse(polygon2._is_inside(polygon1))

        polygon2.invert()

        self.assertTrue(polygon1._is_inside(polygon2))
        self.assertFalse(polygon2._is_inside(polygon1))

        vertices = np.array([[100, 100, 130, 130],
                             [41, 50, 50, 41]]).T

        polygon2 = SphPolygon(np.deg2rad(vertices))

        self.assertFalse(polygon1._is_inside(polygon2))
        self.assertFalse(polygon2._is_inside(polygon1))

        polygon2.invert()

        self.assertTrue(polygon1._is_inside(polygon2))
        self.assertFalse(polygon2._is_inside(polygon1))

        vertices = VERTICES_TEST_IS_INSIDE1

        polygon1 = SphPolygon(np.deg2rad(vertices))

        vertices = VERTICES_TEST_IS_INSIDE2

        polygon2 = SphPolygon(np.deg2rad(vertices))

        self.assertFalse(polygon2._is_inside(polygon1))
        self.assertFalse(polygon1._is_inside(polygon2))

    def test_is_inside_float32(self):
        """Test that precision dependent calculations work.

        Some of the precision math can fail with only 32-bit floats.

        """
        b_verts = np.array([[-1.3440281, 0.12873407],
                            [-1.3714675, 0.17091802],
                            [-1.5773474, 0.14104952],
                            [-1.5913332, 0.09375837],
                            [-1.598127, 0.12360403],
                            [-1.7618076, 0.65335757],
                            [-1.7746785, 0.6819248],
                            [-1.7645605, 0.73133504],
                            [-1.4870182, 0.77076954],
                            [-1.4529741, 0.7284483],
                            [-1.4474869, 0.69782615],
                            [-1.3499607, 0.15860939]], dtype=np.float32)
        t_verts = np.array([[-1.80760407, 0.82227004],
                            [-1.65620456, 0.82549202],
                            [-1.34904288, 0.81162246],
                            [-1.2014294, 0.79483332],
                            [-1.21090379, 0.75978692],
                            [-1.26056844, 0.54041217],
                            [-1.30037964, 0.31453188],
                            [-1.33298796, 0.09364747],
                            [-1.33784689, 0.05810091],
                            [-1.44211119, 0.07002157],
                            [-1.656768, 0.07992753],
                            [-1.76233241, 0.07762154],
                            [-1.7639324, 0.11370358],
                            [-1.77469158, 0.33765409],
                            [-1.78788255, 0.5660454],
                            [-1.8044336, 0.78705058]], dtype=np.float32)
        b_poly = SphPolygon(b_verts)
        t_poly = SphPolygon(t_verts)
        assert b_poly._is_inside(t_poly)

    def test_union_polygons_overlap_partially(self):
        """Test the union method."""
        vertices = np.array([[180, 90, 0, -90],
                             [89, 89, 89, 89]]).T
        poly1 = SphPolygon(np.deg2rad(vertices))
        vertices = np.array([[-45, -135, 135, 45],
                             [89, 89, 89, 89]]).T
        poly2 = SphPolygon(np.deg2rad(vertices))

        uni = np.array([[157.5, 89.23460094],
                        [-225., 89.],
                        [112.5, 89.23460094],
                        [90., 89.],
                        [67.5, 89.23460094],
                        [45., 89.],
                        [22.5, 89.23460094],
                        [0., 89.],
                        [-22.5, 89.23460094],
                        [-45., 89.],
                        [-67.5, 89.23460094],
                        [-90., 89.],
                        [-112.5, 89.23460094],
                        [-135., 89.],
                        [-157.5, 89.23460094],
                        [-180., 89.]])

        poly_union = poly1.union(poly2)

        self.assertTrue(np.allclose(poly_union.vertices, np.deg2rad(uni)))

    def test_union_polygons_overlaps_completely(self):
        """Test the union method when one polygon is entirely inside the other."""
        vertices = np.array([[1, 1, 20, 20],
                             [1, 20, 20, 1]]).T

        poly1 = SphPolygon(np.deg2rad(vertices))

        vertices = np.array([[0, 0, 30, 30],
                             [0, 30, 30, 0]]).T
        poly2 = SphPolygon(np.deg2rad(vertices))

        poly_union1 = poly1.union(poly2)
        poly_union2 = poly2.union(poly1)

        expected = np.deg2rad(np.array([[0, 0, 30, 30],
                                        [0, 30, 30, 0]]).T)

        self.assertTrue(np.allclose(poly_union1.vertices, expected))

        self.assertTrue(np.allclose(poly_union2.vertices, expected))

    def test_intersection(self):
        """Test the intersection function."""
        vertices = np.array([[180, 90, 0, -90],
                             [89, 89, 89, 89]]).T
        poly1 = SphPolygon(np.deg2rad(vertices))
        vertices = np.array([[-45, -135, 135, 45],
                             [89, 89, 89, 89]]).T
        poly2 = SphPolygon(np.deg2rad(vertices))

        inter = np.array([[157.5, 89.23460094],
                          [112.5, 89.23460094],
                          [67.5, 89.23460094],
                          [22.5, 89.23460094],
                          [-22.5, 89.23460094],
                          [-67.5, 89.23460094],
                          [-112.5, 89.23460094],
                          [-157.5, 89.23460094]])
        poly_inter = poly1.intersection(poly2)

        self.assertTrue(np.allclose(poly_inter.vertices,
                                    np.deg2rad(inter)))

        # Test 2 polygons sharing 2 contiguous edges.

        vertices1 = np.array([[-10, 10],
                              [-5, 10],
                              [0, 10],
                              [5, 10],
                              [10, 10],
                              [10, -10],
                              [-10, -10]])

        vertices2 = np.array([[-5, 10],
                              [0, 10],
                              [5, 10],
                              [5, -5],
                              [-5, -5]])

        vertices3 = np.array([[5, 10],
                              [5, -5],
                              [-5, -5],
                              [-5, 10],
                              [0, 10]])

        poly1 = SphPolygon(np.deg2rad(vertices1))
        poly2 = SphPolygon(np.deg2rad(vertices2))
        poly_inter = poly1.intersection(poly2)

        self.assertTrue(np.allclose(poly_inter.vertices,
                                    np.deg2rad(vertices3)))

        # Test when last node of the intersection is the last vertice of the
        # second polygon.

        swath_vertices = np.array([[-115.32268301, 66.32946139],
                                   [-61.48397172, 58.56799254],
                                   [-60.25004314, 58.00754686],
                                   [-71.35057076, 49.60229517],
                                   [-113.746486, 56.03008985]])
        area_vertices = np.array([[-68.32812107, 52.3480829],
                                  [-67.84993896, 53.07015692],
                                  [-55.54651296, 64.9254637],
                                  [-24.63341856, 74.24628796],
                                  [-31.8996363, 27.99907764],
                                  [-39.581043, 37.0639821],
                                  [-50.90185988, 45.56296169],
                                  [-67.43022017, 52.12399581]])

        res = np.array([[-62.77837918, 59.12607053],
                        [-61.48397172, 58.56799254],
                        [-60.25004314, 58.00754686],
                        [-71.35057076, 49.60229517],
                        [-113.746486, 56.03008985],
                        [-115.32268301, 66.32946139]])

        poly1 = SphPolygon(np.deg2rad(swath_vertices))
        poly2 = SphPolygon(np.deg2rad(area_vertices))

        poly_inter = poly1.intersection(poly2)
        self.assertTrue(np.allclose(poly_inter.vertices,
                                    np.deg2rad(res)))

        poly_inter = poly2.intersection(poly1)
        self.assertTrue(np.allclose(poly_inter.vertices,
                                    np.deg2rad(res)))

    def test_consistent_radius(self):
        poly1 = np.array([(-50, 69), (-36, 69), (-36, 64), (-50, 64)])
        poly2 = np.array([(-46, 68), (-40, 68), (-40, 65), (-45, 65)])
        poly_outer = SphPolygon(np.deg2rad(poly1), radius=6371)
        poly_inner = SphPolygon(np.deg2rad(poly2), radius=6371)
        poly_inter = poly_outer.intersection(poly_inner)
        self.assertAlmostEqual(poly_inter.radius, poly_inner.radius)
        # Well, now when we are at it.
        self.assertAlmostEqual(poly_inter.area(), poly_inner.area())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2010-2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Multiprocessing versions of KDTree and Proj classes."""

from __future__ import absolute_import

import ctypes
import multiprocessing as mp

import numpy as np
import pyproj

try:
    import numexpr as ne
except ImportError:
    ne = None

from ._multi_proc import Scheduler, shmem_as_ndarray

# Earth radius
R = 6370997.0


class cKDTree_MP(object):
    """Multiprocessing cKDTree subclass, shared memory."""

    def __init__(self, data, leafsize=10, nprocs=2, chunk=None,
                 schedule='guided'):
        """Prepare shared memory for KDTree operations.

        Same as cKDTree.__init__ except that an internal copy of data to shared memory is made.

        Extra keyword arguments:
        chunk : Minimum chunk size for the load balancer.
        schedule: Strategy for balancing work load
        ('static', 'dynamic' or 'guided').
        """
        self.n, self.m = data.shape
        # Allocate shared memory for data
        self.shmem_data = mp.RawArray(ctypes.c_double, self.n * self.m)

        # View shared memory as ndarray, and copy over the data.
        # The RawArray objects have information about the dtype and
        # buffer size.
        _data = shmem_as_ndarray(self.shmem_data).reshape((self.n, self.m))
        _data[:, :] = data

        # Initialize parent, we must do this last because
        # cKDTree stores a reference to the data array. We pass in
        # the copy in shared memory rather than the origial data.
        self.leafsize = leafsize
        self._nprocs = nprocs
        self._chunk = chunk
        self._schedule = schedule

    def query(self, x, k=1, eps=0, p=2, distance_upper_bound=np.inf):
        """Query for points at index 'x' parallelized with multiple processes and shared memory."""
        # allocate shared memory for x and result
        nx = x.shape[0]
        shmem_x = mp.RawArray(ctypes.c_double, nx * self.m)
        shmem_d = mp.RawArray(ctypes.c_double, nx * k)
        shmem_i = mp.RawArray(ctypes.c_int, nx * k)

        # view shared memory as ndarrays
        _x = shmem_as_ndarray(shmem_x).reshape((nx, self.m))
        if k == 1:
            _d = shmem_as_ndarray(shmem_d)
            _i = shmem_as_ndarray(shmem_i)
        else:
            _d = shmem_as_ndarray(shmem_d).reshape((nx, k))
            _i = shmem_as_ndarray(shmem_i).reshape((nx, k))

        # copy x to shared memory
        _x[:] = x

        # set up a scheduler to load balance the query
        scheduler = Scheduler(nx, self._nprocs, chunk=self._chunk,
                              schedule=self._schedule)

        # query with multiple processes
        query_args = [scheduler, self.shmem_data, self.n, self.m,
                      self.leafsize, shmem_x, nx, shmem_d, shmem_i,
                      k, eps, p, distance_upper_bound]

        _run_jobs(_parallel_query, query_args, self._nprocs)
        # return results (private memory)
        return _d.copy(), _i.copy()


class BaseProj(pyproj.Proj):
    """Helper class for easier backwards compatibility."""

    def __init__(self, projparams=None, preserve_units=True, **kwargs):
        # have to have this because pyproj uses __new__
        # subclasses would fail when calling __init__ otherwise
        super(BaseProj, self).__init__(projparams=projparams,
                                       preserve_units=preserve_units,
                                       **kwargs)


class Proj(BaseProj):
    """Helper class to skip transforming lon/lat projection coordinates."""

    def __call__(self, data1, data2, inverse=False, radians=False,
                 errcheck=False, nprocs=1):
        """Transform coordinates to coordinate system except for geographic coordinate systems."""
        if self.crs.is_geographic:
            return data1, data2
        return super(Proj, self).__call__(data1, data2, inverse=inverse,
                                          radians=radians, errcheck=errcheck)


class Proj_MP(BaseProj):
    """Multi-processing version of the pyproj Proj class."""

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        super(Proj_MP, self).__init__(*args, **kwargs)

    def __call__(self, data1, data2, inverse=False, radians=False,
                 errcheck=False, nprocs=2, chunk=None, schedule='guided'):
        """Transform coordinates to coordinates in the current coordinate system."""
        if self.crs.is_geographic:
            return data1, data2

        grid_shape = data1.shape
        n = data1.size

        # Create shared memory
        shmem_data1 = mp.RawArray(ctypes.c_double, n)
        shmem_data2 = mp.RawArray(ctypes.c_double, n)
        shmem_res1 = mp.RawArray(ctypes.c_double, n)
        shmem_res2 = mp.RawArray(ctypes.c_double, n)

        # view shared memory as ndarrays
        _data1 = shmem_as_ndarray(shmem_data1)
        _data2 = shmem_as_ndarray(shmem_data2)
        _res1 = shmem_as_ndarray(shmem_res1)
        _res2 = shmem_as_ndarray(shmem_res2)

        # copy input data to shared memory
        _data1[:] = data1.ravel()
        _data2[:] = data2.ravel()

        # set up a scheduler to load balance the query
        scheduler = Scheduler(n, nprocs, chunk=chunk, schedule=schedule)

        # Projection with multiple processes
        proj_call_args = [scheduler, shmem_data1, shmem_data2, shmem_res1,
                          shmem_res2, self._args, self._kwargs, inverse,
                          radians, errcheck]

        _run_jobs(_parallel_proj, proj_call_args, nprocs)
        return _res1.copy().reshape(grid_shape), _res2.copy().reshape(grid_shape)


class Cartesian(object):
    """Cartesian coordinates."""

    def __init__(self, *args, **kwargs):
        pass

    def transform_lonlats(self, lons, lats):
        """Transform longitudes and latitues to cartesian coordinates."""
        if np.issubdtype(lons.dtype, np.integer):
            lons = lons.astype(np.float64)
        coords = np.zeros((lons.size, 3), dtype=lons.dtype)
        if ne:
            deg2rad = np.pi / 180  # noqa: F841
            coords[:, 0] = ne.evaluate("R*cos(lats*deg2rad)*cos(lons*deg2rad)")
            coords[:, 1] = ne.evaluate("R*cos(lats*deg2rad)*sin(lons*deg2rad)")
            coords[:, 2] = ne.evaluate("R*sin(lats*deg2rad)")
        else:
            coords[:, 0] = R * np.cos(np.deg2rad(lats)) * np.cos(np.deg2rad(lons))
            coords[:, 1] = R * np.cos(np.deg2rad(lats)) * np.sin(np.deg2rad(lons))
            coords[:, 2] = R * np.sin(np.deg2rad(lats))
        return coords


Cartesian_MP = Cartesian


def _run_jobs(target, args, nprocs):
    """Run process pool."""
    # return status in shared memory
    # access to these values are serialized automatically
    ierr = mp.Value(ctypes.c_int, 0)
    warn_msg = mp.Array(ctypes.c_char, 1024)

    args.extend((ierr, warn_msg))

    pool = [mp.Process(target=target, args=args) for n in range(nprocs)]
    for p in pool:
        p.start()
    for p in pool:
        p.join()
    if ierr.value != 0:
        raise RuntimeError('%d errors in worker processes. Last one reported:\n%s' %
                           (ierr.value, warn_msg.value.decode()))

# This is executed in an external process:


def _parallel_query(scheduler,  # scheduler for load balancing
                    # data needed to reconstruct the kd-tree
                    data, ndata, ndim, leafsize,
                    x, nx, d, i,  # query data and results
                    k, eps, p, dub,  # auxillary query parameters
                    ierr, warn_msg):  # return values (0 on success)

    try:
        # View shared memory as ndarrays.
        _data = shmem_as_ndarray(data).reshape((ndata, ndim))
        _x = shmem_as_ndarray(x).reshape((nx, ndim))
        if k == 1:
            _d = shmem_as_ndarray(d)
            _i = shmem_as_ndarray(i)
        else:
            _d = shmem_as_ndarray(d).reshape((nx, k))
            _i = shmem_as_ndarray(i).reshape((nx, k))

        # Reconstruct the kd-tree from the data.
        import scipy.spatial as sp
        kdtree = sp.cKDTree(_data, leafsize=leafsize)

        # Query for nearest neighbours, using slice ranges,
        # from the load balancer.
        for s in scheduler:
            if k == 1:
                _d[s], _i[s] = kdtree.query(_x[s, :], k=1, eps=eps, p=p,
                                            distance_upper_bound=dub)
            else:
                _d[s, :], _i[s, :] = kdtree.query(_x[s, :], k=k, eps=eps, p=p,
                                                  distance_upper_bound=dub)
    # An error occured, increment the return value ierr.
    # Access to ierr is serialized by multiprocessing.
    except Exception as e:
        ierr.value += 1
        warn_msg.value = str(e).encode()


def _parallel_proj(scheduler, data1, data2, res1, res2, proj_args, proj_kwargs,
                   inverse, radians, errcheck, ierr, warn_msg):
    try:
        # View shared memory as ndarrays.
        _data1 = shmem_as_ndarray(data1)
        _data2 = shmem_as_ndarray(data2)
        _res1 = shmem_as_ndarray(res1)
        _res2 = shmem_as_ndarray(res2)

        # Initialise pyproj
        proj = pyproj.Proj(*proj_args, **proj_kwargs)

        # Reproject data segment
        for s in scheduler:
            _res1[s], _res2[s] = proj(_data1[s], _data2[s], inverse=inverse,
                                      radians=radians, errcheck=errcheck)

    # An error occured, increment the return value ierr.
    # Access to ierr is serialized by multiprocessing.
    except Exception as e:
        ierr.value += 1
        warn_msg.value = str(e).encode()


def _parallel_transform(scheduler, lons, lats, n, coords, ierr, warn_msg):
    try:
        # View shared memory as ndarrays.
        _lons = shmem_as_ndarray(lons)
        _lats = shmem_as_ndarray(lats)
        _coords = shmem_as_ndarray(coords).reshape((n, 3))

        # Transform to cartesian coordinates
        for s in scheduler:
            _coords[s, 0] = R * \
                np.cos(np.radians(_lats[s])) * np.cos(np.radians(_lons[s]))
            _coords[s, 1] = R * \
                np.cos(np.radians(_lats[s])) * np.sin(np.radians(_lons[s]))
            _coords[s, 2] = R * np.sin(np.radians(_lats[s]))
    # An error occured, increment the return value ierr.
    # Access to ierr is serialized by multiprocessing.
    except Exception as e:
        ierr.value += 1
        warn_msg.value = str(e).encode()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Registry of resampler classes."""

from __future__ import annotations

import functools
import warnings
from functools import lru_cache
from typing import Callable, Type

from pyresample._compat import entry_points

from .resampler import Resampler

RESAMPLER_REGISTRY: dict[str, Type[Resampler]] = {}


def register_resampler(resampler_name: str, resampler_cls: Type[Resampler]) -> None:
    """Register :class:`~pyresample.future.resampler.Resampler` subclass for future use.

    Args:
        resampler_name:
            Name of the resampler in the registry. This name can then be used
            in functions like
            :func:`~pyresample.future.resamplers.registry.create_resampler`.
        resampler_cls:
            Subclass of
            :class:`~pyresample.future.resamplers.resampler.Resampler` that
            will be added to the registry.

    Examples:
        Register a custom class::

            register_resampler("my_resampler", MyResamplerClass)

        Register as a plugin from third-party package (in your setup.py)::

            entry_points = {
                "pyresample.resamplers": [
                    "my_resampler = mypkg.mymodule:MyResamplerClass",
                ],
            }

    """
    if resampler_name in RESAMPLER_REGISTRY:
        raise ValueError(
            f"Resampler with name '{resampler_name} is already registered. "
            "Use 'unregister_resampler' to make the name available.")

    RESAMPLER_REGISTRY[resampler_name] = resampler_cls


def unregister_resampler(resampler_name: str) -> None:
    """Remove previously registered Resampler so it can't be used anymore."""
    del RESAMPLER_REGISTRY[resampler_name]


def with_loaded_registry(callable: Callable) -> Callable:
    """Load and verify registry plugins before calling the decorated object.

    Note: This decorator is structured in a way that this plugin loading only
    happens on the usage of the provided callable instead of on import time.

    """
    def _wrapper(*args, **kwargs) -> Callable:
        _load_entry_point_resamplers()
        if not RESAMPLER_REGISTRY:
            warnings.warn("No builtin resamplers found. This probably means you "
                          "installed pyresample in editable mode. Try reinstalling "
                          "pyresample to ensure builtin resamplers are included.")
        return callable(*args, **kwargs)
    return functools.update_wrapper(_wrapper, callable)


@lru_cache(1)
def _load_entry_point_resamplers():
    """Load setuptools plugins via entry_points.

    Based on https://packaging.python.org/guides/creating-and-discovering-plugins/#using-package-metadata.

    """
    discovered_plugins = entry_points(group="pyresample.resamplers")
    for entry_point in discovered_plugins:
        try:
            loaded_resampler = entry_point.load()
        except ImportError:
            warnings.warn(f"Unable to load resampler from plugin: {entry_point.name}")
        else:
            register_resampler(entry_point.name, loaded_resampler)


@with_loaded_registry
def list_resamplers() -> list[str, ...]:
    """Get sorted list of registered resamplers."""
    resampler_names = sorted(RESAMPLER_REGISTRY.keys())
    return resampler_names


@with_loaded_registry
def create_resampler(
        src_geom,
        dst_geom,
        resampler: str = None,
        cache=None,
        **kwargs
) -> Resampler:
    """Create instance of a :class:`~pyresample.future.resampler.Resampler` with the provided arguments.

    Args:
        src_geom:
            Geometry object defining the source geographic region that input
            data lies on and will be resampled from.
        dst_geom:
            Geometry object defining the destination geographic region that
            input data will be resampled to.
        resampler:
            The name of a resampler class that has been previously
            registered with :func:`~pyresample.future.resampler_registry` and
            will be instantiated. If not provided then a registered Resampler
            class will be chosen based on the geometry types provided. This
            is currently always the 'nearest' (nearest neighbor) resampler.
        cache:
            ResampleCache instance used by
            the resampler to cache intermediate results for improved resampling
            performance on multiple executions or future use of the resampler.
        kwargs:
            Additional keyword arguments to pass to the Resampler. Note that
            most resamplers do not have additional keyword arguments on
            creation, but instead have extra arguments passed when their
            ``resample`` methods are called.

    """
    if resampler is None:
        resampler = "nearest"
    resampler_cls = RESAMPLER_REGISTRY[resampler]
    return resampler_cls(src_geom, dst_geom, cache=cache, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Test base resampler class functionality."""
from __future__ import annotations

from unittest import mock

import numpy as np
import pytest
from pytest_lazyfixture import lazy_fixture

from pyresample.future.resamplers.resampler import Resampler


class FakeResampler(Resampler):
    """Helper fake resampler for easier testing."""

    def __init__(self, *args, **kwargs):
        self.precompute = mock.Mock(wraps=self.precompute)
        self.resample = mock.Mock(wraps=self.resample)
        self._internal_cache = {}
        super().__init__(*args, **kwargs)

    def precompute(self, **kwargs):
        hash = self._get_hash(**kwargs)
        if hash not in self._internal_cache:
            some_result = np.empty((2, 2))
            self._internal_cache[hash] = some_result
        return self._internal_cache[hash]

    def resample(self, data, **kwargs):
        self.precompute(**kwargs)
        hash = self._get_hash(**kwargs)
        _ = self._internal_cache[hash]
        return np.empty(self.target_geo_def.shape)


@pytest.mark.parametrize(
    "src",
    [
        lazy_fixture("swath_def_2d_numpy"),
        lazy_fixture("swath_def_2d_dask"),
        lazy_fixture("swath_def_2d_xarray_numpy"),
        lazy_fixture("swath_def_2d_xarray_dask"),
    ]
)
@pytest.mark.parametrize(
    "dst",
    [
        lazy_fixture("area_def_lcc_conus_1km"),
    ]
)
def test_resampler(src, dst):
    """Test basic operations of the base resampler with and without a caching."""
    rs = FakeResampler(src, dst)
    some_data = np.zeros(src.shape, dtype=np.float64)
    resample_results = rs.resample(some_data)
    rs.precompute.assert_called_once()
    assert resample_results.shape == dst.shape



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Load an AreaDefinition object from a netCDF/CF file."""

import pyproj

# list of valid CF grid mappings:
_valid_cf_type_of_grid_mapping = \
    ('albers_conical_equal_area',
     'azimuthal_equidistant',
     'geostationary',
     'lambert_azimuthal_equal_area',
     'lambert_conformal_conic',
     'lambert_cylindrical_equal_area',
     'latitude_longitude',
     'mercator',
     'oblique_mercator',
     'orthographic',
     'polar_stereographic',
     'rotated_latitude_longitude',
     'sinusoidal',
     'stereographic',
     'transverse_mercator',
     'vertical_perspective')

# dictionnary with the standard_names accepted by CF per projection type
#   this can be used for reading from and writing to CF files
_valid_cf_coordinate_standardnames = {}
# specific name for most grid mappings
_valid_cf_coordinate_standardnames['default'] = dict()
_valid_cf_coordinate_standardnames['default']['x'] = ('projection_x_coordinate',)
_valid_cf_coordinate_standardnames['default']['y'] = ('projection_y_coordinate',)
# specific name for the latitude_longitude grid mapping
_valid_cf_coordinate_standardnames['latitude_longitude'] = dict()
_valid_cf_coordinate_standardnames['latitude_longitude']['x'] = ('longitude',)
_valid_cf_coordinate_standardnames['latitude_longitude']['y'] = ('latitude',)
# specific name for the rotated_latitude_longitude grid mapping
_valid_cf_coordinate_standardnames['rotated_latitude_longitude'] = dict()
_valid_cf_coordinate_standardnames['rotated_latitude_longitude']['x'] = ('grid_longitude',)
_valid_cf_coordinate_standardnames['rotated_latitude_longitude']['y'] = ('grid_latitude',)
# specific name for the geostationary grid mapping (we support two flavors)
_valid_cf_coordinate_standardnames['geostationary'] = dict()
_valid_cf_coordinate_standardnames['geostationary']['x'] = (
    'projection_x_angular_coordinate', 'projection_x_coordinate',)
_valid_cf_coordinate_standardnames['geostationary']['y'] = (
    'projection_y_angular_coordinate', 'projection_y_coordinate',)


def _convert_XY_CF_to_Proj(crs, axis_info):
    """Convert XY values from CF to PROJ convention. With CF =< 1.9 only affects geostrationary projection."""
    crs_dict = crs.to_dict()
    axis_units = axis_info.get('unit') or 'radians'  # unit could be None
    if crs_dict['proj'] == 'geos' and axis_units == 'radians':
        # for geostationary projection, the values stored as x/y in CF are not directly
        #  the x/y along the projection axes, but are rather the scanning angles from
        #  the satellite. We must multiply them by the height of the satellite.
        satellite_height = crs_dict['h']
        for k in ('first', 'last', 'spacing'):
            axis_info[k] *= satellite_height
        # the unit is now the default (meters)
        axis_info['units'] = 'm'

    return axis_info


def _load_crs_from_cf_gridmapping(nc_handle, grid_mapping_varname):
    """Initialize a CRS object from a CF grid_mapping variable."""
    # check the variable exists
    try:
        v = nc_handle[grid_mapping_varname]
    except KeyError:
        raise KeyError("Variable '{}' does not exist in netCDF file".format(grid_mapping_varname))

    # check this indeed is a supported grid mapping variable
    try:
        if v.grid_mapping_name not in _valid_cf_type_of_grid_mapping:
            raise ValueError("Not a valid CF grid_mapping variable ({})".format(grid_mapping_varname))
    except AttributeError:
        # no :grid_mapping_name thus it cannot be a valid grid_mapping variable
        raise ValueError("Not a valid CF grid_mapping variable ({})".format(grid_mapping_varname))

    # use pyproj to load the CRS
    return pyproj.CRS.from_cf(v.attrs)


def _is_valid_coordinate_standardname(coord_standard_name, axis, type_of_grid_mapping):
    """Check that a CF coordinate variable has the expected CF standard_name with regard to the typw of grid mapping."""
    valid = False

    if axis not in ('x', 'y'):
        raise ValueError("axis= parameter must be 'x' or 'y'")

    if type_of_grid_mapping != 'default' and type_of_grid_mapping not in _valid_cf_type_of_grid_mapping:
        raise ValueError("grid_mapping_name {} is not a valid CF one".format(type_of_grid_mapping))

    # access the valid standard_names (also handle the 'default')
    try:
        valid_coord_standard_names = _valid_cf_coordinate_standardnames[type_of_grid_mapping][axis]
    except KeyError:
        valid_coord_standard_names = _valid_cf_coordinate_standardnames['default'][axis]

    # test for validity
    valid = coord_standard_name in valid_coord_standard_names

    return valid


def _is_valid_coordinate_variable(nc_handle, coord_varname, axis, type_of_grid_mapping):
    """Check if a variable is a valid CF coordinate variable."""
    if axis not in ('x', 'y'):
        raise ValueError("axis= parameter must be 'x' or 'y'")

    try:
        coord_var = nc_handle[coord_varname]
    except KeyError:
        raise KeyError("Variable '{}' does not exist in netCDF file".format(coord_varname))

    try:
        coord_standard_name = coord_var.standard_name
        valid = _is_valid_coordinate_standardname(coord_standard_name, axis, type_of_grid_mapping)
    except AttributeError:
        # if the coordinate variable is missing a standard_name, it cannot be a valid CF coordinate axis
        valid = False

    return valid


def _load_cf_axis_info(nc_handle, coord_varname):
    """Load and compute information for a coordinate axis (e.g. first & last values, spacing, length, etc...)."""
    # this requires reading the data, we only read first and last
    first = (nc_handle[coord_varname][0]).item()
    last = (nc_handle[coord_varname][-1]).item()
    nb = len(nc_handle[coord_varname])

    # spacing and sign of the axis
    delta = float(last - first) / (nb - 1)
    spacing = abs(delta)
    sign = delta / spacing

    # get the unit information
    try:
        unit = nc_handle[coord_varname].units
    except AttributeError:
        unit = None

    # some units that are valid in CF are not valid to pass to proj
    if unit.startswith('rad') or \
       unit.startswith('deg'):
        unit = None

    # return in a dictionnary structure
    ret = {'first': first, 'last': last, 'spacing': spacing,
           'nb': nb, 'sign': sign, 'unit': unit}

    return ret


def _get_area_extent_from_cf_axis(x, y):
    """Compute the area_extent of the AreaDefinition object from the information on the x and y axes."""
    # find the ll: lower-left and ur: upper-right.
    # x['first'], y['first'] is always the Upper Left corner
    #   (think of numpy's convention for a 2D image with index 0,0 in top left).
    ll_x, ll_y = x['first'], y['last']
    ur_x, ur_y = x['last'], y['first']

    # handle the half-pixel offset between the center of corner cell (what we have in the axis info)
    #   and the corner of corner cell (what AreaDefinition expects)
    ll_x -= x['sign'] * 0.5 * x['spacing']
    ur_x += x['sign'] * 0.5 * x['spacing']
    ll_y += y['sign'] * 0.5 * y['spacing']
    ur_y -= y['sign'] * 0.5 * y['spacing']

    # return as tuple
    ret = (ll_x, ll_y, ur_x, ur_y)

    return ret


def _guess_cf_axis_varname(nc_handle, variable, axis, type_of_grid_mapping):
    """Guess the name of the netCDF variable holding the coordinate axis of a netCDF field."""
    ret = None

    if axis not in ('x', 'y'):
        raise ValueError("axis= parameter must be 'x' or 'y'")

    # the name of y and x are in the dimensions of the variable=
    try:
        dims = nc_handle[variable].dims
    except KeyError:
        raise KeyError("variable {} not found in file".format(variable))

    for dim in dims:
        # test if each dim is a valid CF coordinate variable
        if _is_valid_coordinate_variable(nc_handle, dim, axis, type_of_grid_mapping):
            ret = dim
            break

    return ret


def _guess_cf_lonlat_varname(nc_handle, variable, lonlat):
    """Guess the name of the netCDF variable holding the longitude (or latitude) of a netCDF field."""
    ret = None

    if lonlat not in ('lon', 'lat'):
        raise ValueError("lonlat= parameter must be 'lon' or 'lat'")

    # lat/lon are either directly a dimension, or a :coordinates.

    # By default (decode_cf=True) xarray puts all dims and :coordinates in .coords
    #   and remove the :coordinates attribute
    try:
        search_list = list(nc_handle[variable].coords)
    except KeyError:
        raise KeyError("variable {} not found in file".format(variable))

    # if decode_cf=False was used, the look at the :coordinates attribute
    if 'coordinates' in nc_handle[variable].attrs.keys():
        search_list += (nc_handle[variable].attrs['coordinates']).split()

    # go through the list of variables and check if one of them is lat / lon
    for v in search_list:
        try:
            # this allows for both 'latitude' and 'rotated_latitude'...
            if {'lat': 'latitude', 'lon': 'longitude'}[lonlat] in nc_handle[v].standard_name:
                ret = v
                break
        except AttributeError:
            # no 'standard_name'. this is not what we are looking for.
            pass

    return ret


def _load_cf_area_one_variable_crs(nc_handle, variable):
    """Load the CRS corresponding to variable."""
    grid_mapping_variable = None
    variable_is_itself_gridmapping = False
    # test if the variable has a grid_mapping attribute
    if hasattr(nc_handle[variable], 'grid_mapping'):
        # good. attempt to load the grid_mapping information into a pyproj object
        crs = _load_crs_from_cf_gridmapping(nc_handle, nc_handle[variable].grid_mapping)
        grid_mapping_variable = nc_handle[variable].grid_mapping
    elif hasattr(nc_handle[variable], 'grid_mapping_name') and \
            nc_handle[variable].grid_mapping_name in _valid_cf_type_of_grid_mapping:
        # this looks like a valid grid_mapping variable
        try:
            # try to load it
            crs = _load_crs_from_cf_gridmapping(nc_handle, variable)
            grid_mapping_variable = variable
            variable_is_itself_gridmapping = True
        except pyproj.exceptions.CRSError as ex:
            raise ValueError("ERROR: pyproj didn't manage to load the CRS: {}".format(ex))
    else:
        # fallback position: maybe the variable is on a basic lat/lon grid with no
        #   grid_mapping. Note: there is no default CRS in CF, we choose WGS84
        grid_mapping_variable = "latlon_default"
        crs = pyproj.CRS.from_string('+proj=latlon +datum=WGS84 +ellps=WGS84')

    # return
    return crs, grid_mapping_variable, variable_is_itself_gridmapping


def _load_cf_area_one_variable_axis(nc_handle, variable, type_of_grid_mapping, y=None, x=None):
    """Identidy and load axis x and y."""
    # if y= or x= are None, guess the variable names for the axis
    xy = dict()
    if y is None and x is None:
        for axis in ('x', 'y'):
            xy[axis] = _guess_cf_axis_varname(nc_handle, variable, axis, type_of_grid_mapping)
            if xy[axis] is None:
                raise ValueError("Could not guess the name of the '{}' axis for {}".format(
                    axis, variable))
    else:
        # y= and x= are provided by the caller. Check they are valid CF coordinate variables
        #   The order is always (y,x)
        xy['y'] = y
        xy['x'] = x
        for axis in ('x', 'y'):
            _valid_axis = _is_valid_coordinate_variable(nc_handle, xy[axis], axis, type_of_grid_mapping)
            if not _valid_axis:
                ve = "Variable x='{}' is not a valid CF coordinate variable for the {} axis".format(xy[axis], axis)
                raise ValueError(ve)

    # we now have the names for the x= and y= coordinate variables: load the info of each axis separately
    axis_info = dict()
    for axis in ('x', 'y'):
        axis_info[axis] = _load_cf_axis_info(nc_handle, xy[axis],)

    return xy, axis_info


def _load_cf_area_one_variable_areadef(axis_info, crs, unit, grid_mapping_variable):
    """Prepare the AreaDefinition object."""
    from pyresample import geometry

    # create shape
    shape = (axis_info['y']['nb'], axis_info['x']['nb'])

    # get area extent from the x and y info
    extent = _get_area_extent_from_cf_axis(axis_info['x'], axis_info['y'])

    # finally prepare the AreaDefinition object
    return geometry.AreaDefinition.from_extent(grid_mapping_variable, crs,
                                               shape, extent, units=unit)


def _load_cf_area_one_variable(nc_handle, variable, y=None, x=None):
    """Load the AreaDefinition corresponding to one netCDF variable/field."""
    if variable not in nc_handle.variables.keys():
        raise KeyError("Variable '{}' does not exist in netCDF file".format(variable))

    # the routine always prepares a cf_info
    cf_info = dict()
    cf_info['variable'] = variable

    # Load a CRS object
    # =================
    crs, grid_mapping_variable, variable_is_itself_gridmapping = _load_cf_area_one_variable_crs(nc_handle, variable)

    # the type of grid_mapping (its grid_mapping_name) impacts several aspects of the CF reader
    if grid_mapping_variable == 'latlon_default':
        type_of_grid_mapping = 'latitude_longitude'
    else:
        try:
            type_of_grid_mapping = nc_handle[grid_mapping_variable].grid_mapping_name
        except AttributeError:
            raise ValueError(
                ("Not a valid CF grid_mapping variable ({}):"
                 "it lacks a :grid_mapping_name attribute").format(grid_mapping_variable))

    cf_info['grid_mapping_variable'] = grid_mapping_variable
    cf_info['type_of_grid_mapping'] = type_of_grid_mapping

    # test if we can allow None for y and x
    if variable_is_itself_gridmapping and (y is None or x is None):
        raise ValueError("When variable= points to the grid_mapping variable itself, y= and x= must be provided")

    # identify and load the x/y axis
    # ==============================
    xy, axis_info = _load_cf_area_one_variable_axis(nc_handle, variable, type_of_grid_mapping, y=y, x=x)

    # there are few cases where the x/y values loaded from the CF files cannot be
    #   used directly in pyresample. We need a conversion:
    for axis in ('x', 'y'):
        axis_info[axis] = _convert_XY_CF_to_Proj(crs, axis_info[axis])

    # transfer information on the axis to the cf_info dict()
    for axis in ('x', 'y'):
        cf_info[axis] = dict()
        cf_info[axis]['varname'] = xy[axis]
        for k in axis_info[axis].keys():
            cf_info[axis][k] = axis_info[axis][k]

    # sanity check: we cannot have different units for x and y
    unit = axis_info['x']['unit']
    if axis_info['x']['unit'] != axis_info['y']['unit']:
        raise ValueError("Cannot have different units for 'x' ({}) and 'y' ({}) axis.".format(
            axis_info['x']['unit'], axis_info['y']['unit']))

    # prepare the AreaDefinition object
    # =================================
    area_def = _load_cf_area_one_variable_areadef(axis_info, crs, unit, grid_mapping_variable)

    return area_def, cf_info


def _load_cf_area_several_variables(nc_handle):
    """Load the AreaDefinition corresponding to several netCDF variables/fields."""
    def _indices_unique_AreaDefs(adefs):
        """Find the indices of unique AreaDefinitions in a list."""
        uniqs = dict()
        for i, adef in enumerate(adefs):
            if adef not in uniqs:  # this uses AreaDefinition.__eq__()
                uniqs[adef] = i

        # return only the indices
        return uniqs.values()

    adefs = []
    infos = []

    # go through all the variables
    for v in nc_handle.variables.keys():

        # skip variables that are less than 2D: they cannot
        #   possibly sustain an AreaDefinition
        if nc_handle[v].ndim < 2:
            continue

        try:
            # try and load an AreaDefinition from this variable
            adef, info = _load_cf_area_one_variable(nc_handle, v)
            # store
            adefs.append(adef)
            infos.append(info)
            # break the loop, we have all we need
            break
        except ValueError:
            # this is not a problem: variable v simply doesn't define an AreaDefinition
            continue

    # go through the loaded AreaDefinitions and find the unique ones.
    indices = _indices_unique_AreaDefs(adefs)
    uniq_adefs = [adefs[ui] for ui in indices]
    uniq_infos = [infos[ui] for ui in indices]

    return uniq_adefs, uniq_infos


def load_cf_area(nc_file, variable=None, y=None, x=None):
    """Load an AreaDefinition object from a netCDF/CF file.

    Parameters
    ----------
    nc_file : string or object
        path to a netCDF/CF file, or opened xarray.Dataset object
    variable : string, optional
        name of the variable to load the AreaDefinition from.
        If the variable is not a CF grid_mapping container variable,
        it should be a variable having a :grid_mapping attribute.
        If variable is None the file will be searched for valid CF
        area definitions
    y : string, optional
        name of the variable to use as 'y' axis of the CF area definition
        If y is None an appropriate 'y' axis will be deduced from the CF file
    x : string, optional
        name of the variable to use as 'x' axis of the CF area definition
        If x is None an appropriate 'x' axis will be deduced from the CF file

    Returns
    -------
    are_def, cf_info : geometry.AreaDefinition object, dict
       cf_info holds info about how the AreaDefinition was defined in the CF file.

    """
    import xarray as xr

    # basic check on the default values of the parameters.
    if (x is not None and y is None) or (x is None and y is not None):
        raise ValueError("You must specify either both or none of x= and y=")

    # the nc_file can be either the path to a netCDF/CF file, or directly an opened xarray.Dataset()
    if isinstance(nc_file, xr.Dataset):
        nc_handle = nc_file
    else:
        #   if the path to a file, open the Dataset access to it
        try:
            nc_handle = xr.open_dataset(nc_file)
        except FileNotFoundError as ex:
            raise FileNotFoundError("This file does not exist ({})".format(ex))
        except (OSError, TypeError) as ex:
            raise OSError("This file is probably not a valid netCDF file ({}).".format(ex))

    if variable is None:
        # if the variable=None, we search through all variables
        area_def, cf_info = _load_cf_area_several_variables(nc_handle)
        if len(area_def) == 0:
            raise ValueError("Found no AreaDefinitions in this netCDF/CF file.")
        elif len(area_def) > 1:
            # there were several area_definitions defined in this file. For now bark.
            raise ValueError("The CF file holds several different AreaDefinitions. Use the variable= keyword.")
        else:
            area_def = area_def[0]
            cf_info = cf_info[0]
    else:
        # the variable= is known, call appropriate routine
        try:
            area_def, cf_info = _load_cf_area_one_variable(nc_handle, variable, y=y, x=x, )
        except ValueError as ve:
            raise ValueError("Found no AreaDefinition associated with variable {} ({})".format(variable, ve))

    # also guess the name of the latitude and longitude variables
    for ll in ('lon', 'lat'):
        cf_info[ll] = _guess_cf_lonlat_varname(nc_handle, cf_info['variable'], ll)
        # this can be None, in which case there was no good lat/lon candidate variable
        #   in the file.

    return area_def, cf_info



# pyresample, Resampling of remote sensing image data in python
#
# Copyright (C) 2019  Pyresample developers
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Code for resampling using bucket resampling."""

import logging

import dask.array as da
import numpy as np
import xarray as xr

from pyresample._spatial_mp import Proj

LOG = logging.getLogger(__name__)


class BucketResampler(object):
    """Bucket resampler.

    Bucket resampling is useful for calculating averages and hit-counts
    when aggregating data to coarser scale grids.

    Below are examples how to use the resampler.

    Read data using Satpy.  The resampling can also be done (apart from
    fractions) directly from Satpy, but this demonstrates the direct
    low-level usage.

    >>> from pyresample.bucket import BucketResampler
    >>> from satpy import Scene
    >>> from satpy.resample import get_area_def
    >>> fname = "hrpt_noaa19_20170519_1214_42635.l1b"
    >>> glbl = Scene(filenames=[fname])
    >>> glbl.load(['4'])
    >>> data = glbl['4']
    >>> lons, lats = data.area.get_lonlats()
    >>> target_area = get_area_def('euro4')

    Initialize the resampler

    >>> resampler = BucketResampler(target_area, lons, lats)

    Calculate the sum of all the data in each grid location:

    >>> sums = resampler.get_sum(data)

    Calculate how many values were collected at each grid location:

    >>> counts = resampler.get_count()

    The average can be calculated from the above two results, or directly
    using the helper method:

    >>> average = resampler.get_average(data)

    Calculate fractions of occurrences of different values in each grid
    location.  The data needs to be categorical (in integers), so
    we'll create some categorical data from the brightness temperature
    data that were read earlier.  The data are returned in a
    dictionary with the categories as keys.

    >>> data = da.where(data > 250, 1, 0)
    >>> fractions = resampler.get_fractions(data, categories=[0, 1])
    >>> import matplotlib.pyplot as plt
    >>> plt.imshow(fractions[0]); plt.show()
    """

    def __init__(self, target_area, source_lons, source_lats):
        self.target_area = target_area
        self.source_lons = source_lons
        self.source_lats = source_lats
        self.prj = Proj(self.target_area.proj_dict)
        self.x_idxs = None
        self.y_idxs = None
        self.idxs = None
        self._get_indices()
        self.counts = None

    def _get_proj_coordinates(self, lons, lats):
        """Calculate projection coordinates.

        Parameters
        ----------
        lons : Numpy or Dask array
            Longitude coordinates
        lats : Numpy or Dask array
            Latitude coordinates
        """
        proj_x, proj_y = self.prj(lons, lats)
        return np.stack((proj_x, proj_y))

    def _get_indices(self):
        """Calculate projection indices.

        Returns
        -------
        x_idxs : Dask array
            X indices of the target grid where the data are put
        y_idxs : Dask array
            Y indices of the target grid where the data are put
        """
        LOG.info("Determine bucket resampling indices")

        # Transform source lons/lats to target projection coordinates x/y
        lons = self.source_lons.ravel()
        lats = self.source_lats.ravel()
        result = da.map_blocks(self._get_proj_coordinates, lons, lats,
                               new_axis=0, chunks=(2,) + lons.chunks)
        proj_x = result[0, :]
        proj_y = result[1, :]

        # Calculate array indices. Orient so that 0-meridian is pointing down.
        adef = self.target_area
        x_res, y_res = adef.resolution
        x_idxs = da.floor((proj_x - adef.area_extent[0]) / x_res).astype(np.int64)
        y_idxs = da.floor((adef.area_extent[3] - proj_y) / y_res).astype(np.int64)

        # Get valid index locations
        mask = (x_idxs >= 0) & (x_idxs < adef.width) & (y_idxs >= 0) & (y_idxs < adef.height)
        self.y_idxs = da.where(mask, y_idxs, -1)
        self.x_idxs = da.where(mask, x_idxs, -1)

        # Convert X- and Y-indices to raveled indexing
        target_shape = self.target_area.shape
        self.idxs = self.y_idxs * target_shape[1] + self.x_idxs

    def get_sum(self, data, skipna=True):
        """Calculate sums for each bin with drop-in-a-bucket resampling.

        Parameters
        ----------
        data : Numpy or Dask array
            Data to be binned and summed.
        skipna : boolean (optional)
                If True, skips NaN values for the sum calculation
                (similarly to Numpy's `nansum`). Buckets containing only NaN are set to zero.
                If False, sets the bucket to NaN if one or more NaN values are present in the bucket
                (similarly to Numpy's `sum`).
                In both cases, empty buckets are set to 0.
                Default: True

        Returns
        -------
        data : Numpy or Dask array
            Bin-wise sums in the target grid
        """
        LOG.info("Get sum of values in each location")

        if isinstance(data, xr.DataArray):
            data = data.data
        data = data.ravel()

        # Remove NaN values from the data when used as weights
        weights = da.where(np.isnan(data), 0, data)

        # Rechunk indices to match the data chunking
        if weights.chunks != self.idxs.chunks:
            self.idxs = da.rechunk(self.idxs, weights.chunks)

        # Calculate the sum of the data falling to each bin
        out_size = self.target_area.size
        sums, _ = da.histogram(self.idxs, bins=out_size, range=(0, out_size),
                               weights=weights, density=False)

        # TODO remove following line in favour of weights = data when dask histogram bug (issue #6935) is fixed
        sums = self._mask_bins_with_nan_if_not_skipna(skipna, data, out_size, sums)

        return sums.reshape(self.target_area.shape)

    def _mask_bins_with_nan_if_not_skipna(self, skipna, data, out_size, statistic):
        if not skipna:
            nans = np.isnan(data)
            nan_bins, _ = da.histogram(self.idxs[nans], bins=out_size,
                                       range=(0, out_size))
            statistic = da.where(nan_bins > 0, np.nan, statistic)
        return statistic

    def _call_pandas_groupby_statistics(self, scipy_method, data, fill_value=None, skipna=None):
        """Calculate statistics (min/max) for each bin with drop-in-a-bucket resampling."""
        import dask.dataframe as dd
        import pandas as pd

        if isinstance(data, xr.DataArray):
            data = data.data
        data = data.ravel()

        # Remove NaN values from the data when used as weights
        weights = da.where(np.isnan(data), 0, data)

        # Rechunk indices to match the data chunking
        if weights.chunks != self.idxs.chunks:
            self.idxs = da.rechunk(self.idxs, weights.chunks)

        # Calculate the min of the data falling to each bin
        out_size = self.target_area.size

        # merge into one Dataframe
        df = dd.concat([dd.from_dask_array(self.idxs), dd.from_dask_array(weights)],
                       axis=1)
        df.columns = ['x', 'values']

        if scipy_method == 'min':
            statistics = df.map_partitions(lambda part: part.groupby(
                                           np.digitize(part.x,
                                                       bins=np.linspace(0, out_size, out_size)
                                                       )
                                           )['values'].min())

        elif scipy_method == 'max':
            statistics = df.map_partitions(lambda part: part.groupby(
                                           np.digitize(part.x,
                                                       bins=np.linspace(0, out_size, out_size)
                                                       )
                                           )['values'].max())

        # fill missed index
        statistics = (statistics + pd.Series(np.zeros(out_size))).fillna(0)

        counts = self.get_sum(np.logical_not(np.isnan(data)).astype(np.int64)).ravel()

        # TODO remove following line in favour of weights = data when dask histogram bug (issue #6935) is fixed
        statistics = self._mask_bins_with_nan_if_not_skipna(skipna, data, out_size, statistics)

        # set bin without data to fill value
        statistics = da.where(counts == 0, fill_value, statistics)

        return statistics.reshape(self.target_area.shape)

    def get_min(self, data, fill_value=np.nan, skipna=True):
        """Calculate minimums for each bin with drop-in-a-bucket resampling.

        .. warning::

            The slow :meth:`pandas.DataFrame.groupby` method is temporarily used here,
            as the `dask_groupby <https://github.com/dcherian/dask_groupby>`_ is still under development.

        Parameters
        ----------
        data : Numpy or Dask array
            Data to be binned.
        skipna : boolean (optional)
                If True, skips NaN values for the minimum calculation
                (similarly to Numpy's `nanmin`). Buckets containing only NaN are set to zero.
                If False, sets the bucket to NaN if one or more NaN values are present in the bucket
                (similarly to Numpy's `min`).
                In both cases, empty buckets are set to 0.
                Default: True

        Returns
        -------
        data : Numpy or Dask array
            Bin-wise minimums in the target grid
        """
        LOG.info("Get min of values in each location")
        return self._call_pandas_groupby_statistics('min', data, fill_value, skipna)

    def get_max(self, data, fill_value=np.nan, skipna=True):
        """Calculate maximums for each bin with drop-in-a-bucket resampling.

        .. warning::

            The slow :meth:`pandas.DataFrame.groupby` method is temporarily used here,
            as the `dask_groupby <https://github.com/dcherian/dask_groupby>`_ is still under development.

        Parameters
        ----------
        data : Numpy or Dask array
            Data to be binned.
        skipna : boolean (optional)
                If True, skips NaN values for the maximum calculation
                (similarly to Numpy's `nanmax`). Buckets containing only NaN are set to zero.
                If False, sets the bucket to NaN if one or more NaN values are present in the bucket
                (similarly to Numpy's `max`).
                In both cases, empty buckets are set to 0.
                Default: True

        Returns
        -------
        data : Numpy or Dask array
            Bin-wise maximums in the target grid
        """
        LOG.info("Get max of values in each location")
        return self._call_pandas_groupby_statistics('max', data, fill_value, skipna)

    def get_count(self):
        """Count the number of occurrences for each bin using drop-in-a-bucket resampling.

        Returns
        -------
        data : Dask array
            Bin-wise count of hits for each target grid location
        """
        LOG.info("Get number of values in each location")

        out_size = self.target_area.size

        # Calculate the sum of the data falling to each bin
        if self.counts is None:
            counts, _ = da.histogram(self.idxs, bins=out_size,
                                     range=(0, out_size))
            self.counts = counts.reshape(self.target_area.shape)

        return self.counts

    def get_average(self, data, fill_value=np.nan, skipna=True):
        """Calculate bin-averages using bucket resampling.

        Parameters
        ----------
        data : Numpy or Dask array
            Data to be binned and averaged.
        fill_value : float
            Fill value to mark missing/invalid values in the input data,
            as well as in the binned and averaged output data.
            Default: np.nan
        skipna : bool
            If True, skips missing values (as marked by NaN or `fill_value`) for the average calculation
            (similarly to Numpy's `nanmean`). Buckets containing only missing values are set to fill_value.
            If False, sets the bucket to fill_value if one or more missing values are present in the bucket
            (similarly to Numpy's `mean`).
            In both cases, empty buckets are set to NaN.
            Default: True

        Returns
        -------
        average : Dask array
            Binned and averaged data.
        """
        LOG.info("Get average value for each location")

        if not np.isnan(fill_value):
            data = da.where(data == fill_value, np.nan, data)

        sums = self.get_sum(data, skipna=skipna)
        counts = self.get_sum(np.logical_not(np.isnan(data)).astype(np.int64))

        average = sums / da.where(counts == 0, np.nan, counts)
        average = da.where(np.isnan(average), fill_value, average)

        return average

    def get_fractions(self, data, categories=None, fill_value=np.nan):
        """Get fraction of occurrences for each given categorical value.

        Parameters
        ----------
        data : Numpy or Dask array
            Categorical data to be processed
        categories : iterable or None
            One dimensional list of categories in the data, or None.  If None,
            categories are determined from the data by fully processing the
            data and finding the unique category values.
        fill_value : float
            Fill value to replace missing values.  Default: np.nan
        """
        if categories is None:
            LOG.warning("No categories given, need to compute the data.")
            # compute any dask arrays by converting to numpy
            categories = np.asarray(np.unique(data))
        try:
            num = categories.size
        except AttributeError:
            num = len(categories)
        LOG.info("Get fractions for %d categories", num)
        results = {}
        counts = self.get_count()
        counts = counts.astype(float)
        # Disable logging for calls to get_sum()
        LOG.disabled = True
        for cat in categories:
            cat_data = da.where(data == cat, 1.0, 0.0)

            sums = self.get_sum(cat_data)
            result = sums.astype(float) / counts
            result = da.where(counts == 0.0, fill_value, result)
            results[cat] = result
        # Re-enable logging
        LOG.disabled = False

        return results


def round_to_resolution(arr, resolution):
    """Round the values in *arr* to closest resolution element.

    Parameters
    ----------
    arr : list, tuple, Numpy or Dask array
        Array to be rounded
    resolution : float
        Resolution unit to which data are rounded

    Returns
    -------
    data : Numpy or Dask array
        Source data rounded to the closest resolution unit
    """
    if isinstance(arr, (list, tuple)):
        arr = np.array(arr)
    return resolution * np.round(arr / resolution)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019-2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Base resampler class made for subclassing."""
from __future__ import annotations

import abc
import hashlib
import json
import logging
from typing import Optional, Union

try:
    import xarray as xr
except ImportError:
    xr = None

from pyresample.geometry import AreaDefinition, CoordinateDefinition

logger = logging.getLogger(__name__)
HashType = hashlib._hashlib.HASH


def hash_dict(the_dict: dict, existing_hash: Optional[HashType] = None) -> HashType:
    """Calculate a hash for a dictionary and optionally update an existing hash."""
    if existing_hash is None:
        existing_hash = hashlib.sha1()
    existing_hash.update(json.dumps(the_dict, sort_keys=True).encode('utf-8'))
    return existing_hash


def hash_resampler_geometries(source_geo_def, target_geo_def, **kwargs) -> str:
    """Get hash for the geometries used by a resampler with extra *kwargs*."""
    resampler_hash = source_geo_def.update_hash()
    resampler_hash = target_geo_def.update_hash(resampler_hash)
    resampler_hash = hash_dict(kwargs, resampler_hash)
    return resampler_hash.hexdigest()


def _data_arr_needs_xy_coords(data_arr, area):
    coords_exist = 'x' in data_arr.coords and 'y' in data_arr.coords
    no_xy_dims = 'x' not in data_arr.dims or 'y' not in data_arr.dims
    has_proj_vectors = hasattr(area, 'get_proj_vectors')
    return not (coords_exist or no_xy_dims or not has_proj_vectors)


def _add_xy_units(crs, x_attrs, y_attrs):
    if crs is not None:
        units = crs.axis_info[0].unit_name
        # fix udunits/CF standard units
        units = units.replace('metre', 'meter')
        if units == 'degree':
            y_attrs['units'] = 'degrees_north'
            x_attrs['units'] = 'degrees_east'
        else:
            y_attrs['units'] = units
            x_attrs['units'] = units


def add_xy_coords(data_arr, area, crs=None):
    """Assign x/y coordinates to DataArray from provided area.

    If 'x' and 'y' coordinates already exist then they will not be added.

    Args:
        data_arr (xarray.DataArray): data object to add x/y coordinates to
        area (pyresample.geometry.AreaDefinition): area providing the
            coordinate data.
        crs (pyproj.crs.CRS or None): CRS providing additional information
            about the area's coordinate reference system if available.
            Requires pyproj 2.0+.

    Returns (xarray.DataArray): Updated DataArray object

    """
    if not _data_arr_needs_xy_coords(data_arr, area):
        return data_arr

    x, y = area.get_proj_vectors()
    # convert to DataArrays
    y_attrs = {}
    x_attrs = {}
    _add_xy_units(crs, x_attrs, y_attrs)
    y = xr.DataArray(y, dims=('y',), attrs=y_attrs)
    x = xr.DataArray(x, dims=('x',), attrs=x_attrs)
    return data_arr.assign_coords(y=y, x=x)


def _find_and_assign_crs(data_arr, area):
    # add CRS object if pyproj 2.0+
    try:
        from pyproj import CRS
    except ImportError:
        logger.debug("Could not add 'crs' coordinate with pyproj<2.0")
        crs = None
    else:
        # default lat/lon projection
        latlon_proj = "+proj=latlong +datum=WGS84 +ellps=WGS84"
        # otherwise get it from the area definition
        if hasattr(area, 'crs'):
            crs = area.crs
        else:
            proj_str = getattr(area, 'proj_str', latlon_proj)
            crs = CRS.from_string(proj_str)
        data_arr = data_arr.assign_coords(crs=crs)
    return data_arr, crs


def _update_swath_lonlat_attrs(area):
    # add lon/lat arrays for swath definitions
    # SwathDefinitions created by Satpy should be assigning DataArray
    # objects as the lons/lats attributes so use those directly to
    # maintain original .attrs metadata (instead of converting to dask
    # array).
    lons = area.lons
    lats = area.lats
    if lons.ndim > 2:
        return
    if not isinstance(lons, xr.DataArray):
        dims = ('y', 'x') if lons.ndim == 2 else ('y',)
        lons = xr.DataArray(lons, dims=dims)
        lats = xr.DataArray(lats, dims=dims)
    lons.attrs.setdefault('standard_name', 'longitude')
    lons.attrs.setdefault('long_name', 'longitude')
    lons.attrs.setdefault('units', 'degrees_east')
    lats.attrs.setdefault('standard_name', 'latitude')
    lats.attrs.setdefault('long_name', 'latitude')
    lats.attrs.setdefault('units', 'degrees_north')
    # See https://github.com/pydata/xarray/issues/3068
    # data_arr = data_arr.assign_coords(longitude=lons, latitude=lats)


def add_crs_xy_coords(data_arr, area):
    """Add :class:`pyproj.crs.CRS` and x/y or lons/lats to coordinates.

    For SwathDefinition or GridDefinition areas this will add a
    `crs` coordinate and coordinates for the 2D arrays of `lons` and `lats`.

    For AreaDefinition areas this will add a `crs` coordinate and the
    1-dimensional `x` and `y` coordinate variables.

    Args:
        data_arr (xarray.DataArray): DataArray to add the 'crs'
            coordinate.
        area (pyresample.geometry.AreaDefinition): Area to get CRS
            information from.

    """
    data_arr, crs = _find_and_assign_crs(data_arr, area)

    # Add x/y coordinates if possible
    if isinstance(area, CoordinateDefinition):
        _update_swath_lonlat_attrs(area)
    else:
        # Gridded data (AreaDefinition/StackedAreaDefinition)
        data_arr = add_xy_coords(data_arr, area, crs=crs)
    return data_arr


def update_resampled_coords(old_data, new_data, new_area):
    """Add coordinate information to newly resampled DataArray.

    Args:
        old_data (xarray.DataArray): Old data before resampling.
        new_data (xarray.DataArray): New data after resampling.
        new_area (pyresample.geometry.BaseDefinition): Area definition
            for the newly resampled data.

    """
    # copy over other non-x/y coordinates
    # this *MUST* happen before we set 'crs' below otherwise any 'crs'
    # coordinate in the coordinate variables we are copying will overwrite the
    # 'crs' coordinate we just assigned to the data
    ignore_coords = ('y', 'x', 'crs')
    new_coords = {}
    for cname, cval in old_data.coords.items():
        # we don't want coordinates that depended on the old x/y dimensions
        has_ignored_dims = any(dim in cval.dims for dim in ignore_coords)
        if cname in ignore_coords or has_ignored_dims:
            continue
        new_coords[cname] = cval
    new_data = new_data.assign_coords(**new_coords)

    # add crs, x, and y coordinates
    new_data = add_crs_xy_coords(new_data, new_area)
    # make sure the new area is assigned to the attributes
    new_data.attrs['area'] = new_area
    return new_data


class Resampler:
    """Base abstract resampler class."""

    def __init__(self,
                 source_geo_def: Union[CoordinateDefinition, AreaDefinition],
                 target_geo_def: Union[CoordinateDefinition, AreaDefinition],
                 cache=None,
                 ):
        """Initialize resampler with geolocation information.

        Args:
            source_geo_def:
                Geolocation definition for the data to be resampled
            target_geo_def:
                Geolocation definition for the area to resample data to.
            cache:
                :class:`~pyresample.cache.ResampleCache` instance used by
                the resampler to cache.

        """
        self.source_geo_def = source_geo_def
        self.target_geo_def = target_geo_def
        if isinstance(cache, str):
            # TODO: convenience for file based caching
            raise NotImplementedError()
        self.cache = cache

    @property
    @abc.abstractmethod
    def version(self) -> str:
        """Get version number string of the current resampler for hashing and caching.

        In cases where a resampler has changed enough that past cached files
        are not usable then this version number will be incremented to
        distinguish between the cache entries.

        """

    def _get_hash(self, **kwargs) -> str:
        """Get hash for the current resampler with the given *kwargs*."""
        kwargs.update({
            "resampler_version": self.version,
            "resampler_name": self.__class__.__name__,
        })
        return hash_resampler_geometries(self.source_geo_def, self.target_geo_def, **kwargs)

    def precompute(self, **kwargs):
        """Do the precomputation.

        This is an optional step if the subclass wants to implement more
        complex features like caching or can share some calculations
        between multiple datasets to be processed.

        """
        return None

    def resample(self, data):
        """Resample input ``data`` from the source geometry to the target geometry.

        Args:
            data (ArrayLike): Data to be resampled

        Returns (ArrayLike): Data resampled to the target area

        """
        self.precompute()
        raise NotImplementedError("This method must be implemented by a subclass.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Pyresample developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Functions to support the calculation of a coverage of an area by a set of spherical polygons.

It can for instance be a set of satellite overpasses to be received of a given
local stations over a certain time window where we want to calculate how much
of an area is covered by the onboard scanning instrument(s).
"""

import itertools


class GetNonOverlapUnionsBaseClass():
    """Base class to get the smallest set of union objects that does not overlap.

    The objects are here Python sets of integers - but are abstracts for
    geometrical shapes on a sphere.

    """

    def __init__(self, geom_objects):
        self.geometries = geom_objects
        self.geom_numbers = range(len(geom_objects))

        self._geoms = dict(enumerate(geom_objects))

    def merge(self):
        """Merge all overlapping objects (sets or polygons)."""
        all_unions = self._merge_unions(self._geoms)
        check_keys_int_or_tuple(all_unions)

        self._geoms = {}
        for key, value in all_unions.items():
            if isinstance(key, int):
                self._geoms[key] = value
            else:
                newkey = merge_tuples(key)
                self._geoms[newkey] = value

    def get_polygons(self):
        """Get a list of all non-overlapping polygon unions."""
        return list(self._geoms.values())

    def get_ids(self):
        """Get a list of identifiers identifying the gemoetry objects in each polygon union."""
        return list(self._geoms.keys())

    def _overlaps(self, set1, set2):
        """Check if the two sets overlap each other (have anything in common)."""
        if set1 != set1.difference(set2):
            return True

        return False

    def _find_union_pair(self, geoms):
        """From a set of geometries find a pair that overlaps.

        *geoms* is here expected to be a numbered dict with SphPolygon
        polygons. If no pair of two polygons overlap (that is the union
        returns something different from None or False) then return None.

        Strictly the geometries/objects does not need to be a SphPolygon. The
        only requirement is that it has a union method with the same behaviour.

        """
        if len(geoms) == 1:
            return None

        for id_, komb_pair in zip(itertools.combinations(geoms.keys(), 2),
                                  itertools.combinations(geoms.values(), 2)):
            if self._overlaps(komb_pair[0], komb_pair[1]):
                return id_, komb_pair[0].union(komb_pair[1])

        return None

    def _merge_unions(self, geoms):
        """Merge all overlapping geometry unions.

        Go through the dictionary of geometries and merge overlapping ones until
        there is no change anymore. That is, the input dict of geometries is the
        same as the output:

        """
        retv = self._find_union_pair(geoms)
        if retv is None:
            return geoms

        cc_poly = geoms.copy()

        for idx in [0, 1]:
            del cc_poly[retv[0][idx]]
        cc_poly[retv[0]] = retv[1]

        return self._merge_unions(cc_poly)


class GetNonOverlapUnions(GetNonOverlapUnionsBaseClass):
    """NonOverlapUnions class."""

    def __init__(self, polygons):
        """Init the GetNonOverlapUnions."""
        super(GetNonOverlapUnions, self).__init__(polygons)

    def _overlaps(self, polygon1, polygon2):
        """Check if two polygons overlap each other (have anything in common).

        Return True if they do overlap, otherwise False.

        *polygon1* and *polygon2* are here expected to be SphPolygon
        polygons. Strictly the two input objects do not need to be of type
        SphPolygon. The only requirement is that they have a union method with
        the same behaviour.

        """
        return check_if_two_polygons_overlap(polygon1, polygon2)


def merge_tuples(atuple):
    """Take a nested tuple of integers and concatenate it to a tuple of integers."""
    if not isinstance(atuple, tuple):
        raise TypeError("Function argument must be a tuple!")

    while True:
        # Test if all items in tuple are scalars and can be summed:
        try:
            _ = sum(atuple)
        except TypeError:
            pass
        else:
            return atuple

        atuple = _int_items_to_tuples(atuple)

        try:
            atuple = sum(atuple, ())
        except TypeError:
            return atuple


def _int_items_to_tuples(mytuple):
    """Turn integer scalars in a tuple into tuples."""
    newtup = []
    for item in mytuple:
        if isinstance(item, int):
            newtup.append((item,))
        else:
            newtup.append(item)

    return tuple(newtup)


def check_keys_int_or_tuple(adict):
    """Check if the dictionary keys are integers or tuples.

    If they are not, raise a KeyError
    """
    for key in adict:
        if not isinstance(key, (int, tuple)):
            raise KeyError("Key must be integer or a tuple (of integers)")


def check_if_two_polygons_overlap(polygon1, polygon2):
    """Check if two SphPolygons overlaps."""
    if polygon1.union(polygon2):
        return True

    return False
